from typing import Union, Any, Tuple, Optional
from PyQt6.QtWidgets import (
    QDialog,
    QVBoxLayout,
    QHBoxLayout,
    QGridLayout,
    QDialogButtonBox,
    QComboBox,
    QLineEdit,
    QLabel,
    QWidget,
    QTabWidget,
    QListWidget,
    QPushButton,
    QTableView,
    QCheckBox,
    QListWidgetItem,
)
from PyQt6.QtCore import Qt, pyqtSlot, QEvent
from PyQt6.QtGui import QIntValidator, QDoubleValidator
from .ygModel import (
    unicode_cat_names,
    reverse_unicode_cat_names,
    ygMasters,
    random_id,
    ygFont,
    ygGlyph,
    ygPoint,
    ygcvt
)
from .ygPreferences import ygPreferences

NEW_CV_NAME = "New_Control_Value"
NEW_CV_CONTENT = {"val": 0, "axis": "y", "type": "pos"}

#
# A dialog (makeCVDialog) for creating a CV based on one or two selected points.
#
# A "Font Info" window (fontInfoWindow) for creating/editing/deleting CVs and
# their various properties; editing variation masters; and managing the defaults
# for the current font.
#
# Here is the structure of the Font Info window:
#
# fontInfoWindow ---|
#                   |---- cvEditPane -----|
#                   |                     | --- QListWidget
#                   |                     | --- cvWidget ---|
#                   |                                       |--- general_tab
#                   |                                       |--- same-as tab
#                   |                                       |--- deltas tab
#                   |                                       |--- variants tab
#                   |
#                   |--- mastersWidget ---|
#                   |                     |--- QListWidget
#                   |                     |--- masterWidget
#                   |                     |--- generate variant CVs button
#                   |
#                   |--- font defaults ---|
#                                         |--- Rounding
#                                         |--- Miscellaneous defaults


class cvSource:
    """Mixin superclass for objects that serve CV data."""

    def send_error_message(self, d: dict):
        ...

    def cvt(self):
        ...

    def current_cv(self):
        ...

    def current_cv_name(self):
        ...

    def set_cv_name(self, s: str):
        ...

    def from_current_cv(self, s: str):
        ...

    def set_in_current_cv(self, k: str, s, fallback: Any):
        ...

    def has_key(self, k: str):
        ...

    def del_key(self, k: str):
        ...


class cvEditPane(QWidget, cvSource):
    """A widget with a list of CVs and a cvWidget. Click in the
    list to display and edit that CV.

    params:

    owner: The owner of this widget.

    yg_font (ygFont): The font being edited.

    preferences (ygPreferences): The preferences for this app.

    """

    def __init__(self, owner, yg_font: ygFont, preferences: ygPreferences) -> None:
        super().__init__()
        self.owner = owner
        self.yg_font = yg_font
        self._cvt = self.yg_font.cvt
        self.preferences = preferences
        self.layout_obj = QHBoxLayout()

        # Set up CV list.

        self.cv_list_layout = QVBoxLayout()
        self.button_layout = QHBoxLayout()

        self.cv_list = QListWidget()
        self.cv_list.addItems(self.sorted_cv_list)
        self.current_list_item = self.cv_list.item(0)
        self.cv_list.setCurrentItem(self.current_list_item)
        self._current_cv_name = self.current_list_item.text()
        self.cv_list.itemActivated.connect(self.new_item)

        # And the edit pane.

        self._current_cv = self._cvt.get_cv(self._current_cv_name)
        self.edit_pane = cvWidget(self, self.yg_font, self)

        # Put them together

        self.cv_list_layout.addWidget(self.cv_list)
        add_button = QPushButton("Add")
        del_button = QPushButton("Delete")
        self.button_layout.addWidget(add_button)
        self.button_layout.addWidget(del_button)
        add_button.clicked.connect(self.add_cv)
        del_button.clicked.connect(self.del_cv)
        self.cv_list_layout.addLayout(self.button_layout)
        self.layout_obj.addLayout(self.cv_list_layout)
        self.layout_obj.addWidget(self.edit_pane)
        self.setLayout(self.layout_obj)

    @property
    def sorted_cv_list(self) -> list:
        return sorted(self._cvt.keys, key=lambda s: s.casefold())

    def send_error_message(self, d: dict) -> None:
        self.yg_font.send_error_message(d)

    def add_cv(self) -> None:
        self._current_cv_name = NEW_CV_NAME
        self._cvt.add_cv(self._current_cv_name, NEW_CV_CONTENT)
        self.cv_list.addItem(self._current_cv_name)
        matches = self.cv_list.findItems(
            self._current_cv_name, Qt.MatchFlag.MatchExactly
        )
        if len(matches) > 0:
            self.current_list_item = matches[0]
            self.cv_list.setCurrentItem(self.current_list_item)
            self.new_item(self.current_list_item, forced=True)

    def change_name_in_list(self, n) -> None:
        self.current_list_item.setText(n)

    def del_cv(self) -> None:
        self._cvt.del_cv(self._current_cv_name)
        self.cv_list.clear()
        try:
            self._current_cv_name = list(sorted(self._cvt.keys, key=lambda s: s.casefold()))[0]
        except IndexError:
            return
        self.refresh()

    def refresh(self) -> None:
        """This is the place to figure out whether the source
        or masters have been changed: if not, we don't have to
        go through all this.
        """
        if not len(self._cvt):
            self.add_cv()
        self._current_cv = self._cvt.get_cv(self._current_cv_name)
        self.cv_list.clear()
        self.cv_list.addItems(sorted(self._cvt.keys, key=lambda s: s.casefold()))
        matches = self.cv_list.findItems(
            self._current_cv_name, Qt.MatchFlag.MatchExactly
        )
        if len(matches) > 0:
            self.cv_list.setCurrentItem(matches[0])
        else:
            try:
                current_item = self.cv_list.item(0)
                self.cv_list.setCurrentItem(current_item)
                self._current_cv_name = current_item.text()
                self._current_cv = self._cvt.get_cv(self._current_cv_name)
            except Exception:
                pass
        self.edit_pane.refresh(self)

    def fixup(self) -> None:
        self.edit_pane.fixup()

    def new_item(self, list_item, forced: bool = False) -> None:
        """Switch the view to another cv. Simply delete the
        old cv editing pane and create a new one to put
        in its place.
        """
        new_cv_name = list_item.text()
        if forced or new_cv_name != self._current_cv_name:
            old_pane = self.layout_obj.itemAt(1)
            self.layout_obj.removeItem(old_pane)
            old_pane.widget().deleteLater()
            self._current_cv_name = new_cv_name
            self._current_cv = self._cvt.get_cv(self._current_cv_name)
            self.edit_pane = cvWidget(self, self.yg_font, self)
            self.layout_obj.addWidget(self.edit_pane)

    def cvt(self) -> ygcvt:
        return self._cvt

    def current_cv(self) -> Union[int, dict]:
        return self._current_cv

    def current_cv_name(self) -> str:
        return self._current_cv_name

    def set_cv_name(self, s: str) -> None:
        self._current_cv_name = s

    def from_current_cv(self, s: str):
        try:
            return self._current_cv[s] # type: ignore
        except KeyError:
            return None

    def set_in_current_cv(self, k: str, s: Any, fallback = Optional[str]) -> None:
        if s == "None" or s == "" or s == None:
            if fallback != None:
                self._cvt.set_cv_property(self.current_cv_name(), k, fallback)
            else:
                if k in self._current_cv: # type: ignore
                    self._cvt.del_cv(k)
        else:
            self._cvt.set_cv_property(self.current_cv_name(), k, s)

    def has_key(self, k: str) -> bool:
        return k in self._current_cv # type: ignore

    def del_key(self, k: str) -> None:
        self._cvt.del_cv_property(self.current_cv_name(), k)

    def rename_cv(self, old_name: str, new_name: str) -> None:
        self._cvt.rename(old_name, new_name)

    def showEvent(self, event) -> None:
        self.refresh()

    def hideEvent(self, event) -> None:
        self.fixup()


class fontInfoWindow(QWidget):
    """A one-stop shop for font-level settings: CVs, masters, font-wide
    defaults.

    params:

    yg_font (ygFont): The font being edited.

    preferences (ygPreferences): The preferences for this app.

    """

    def __init__(self, yg_font: ygFont, preferences: ygPreferences) -> None:
        super().__init__()
        self.yg_font = yg_font
        self.cvt = self.yg_font.cvt
        self.preferences = preferences
        self.layout_obj = QVBoxLayout()

        # Set up tabs

        self.tabs = QTabWidget()
        self.cv_tab = cvEditPane(self, self.yg_font, self.preferences)
        self._empty_string = "{}\n"
        self.masters_tab = None
        if self.yg_font.is_variable_font:
            self.masters_tab = mastersWidget(self, self.yg_font)
        self.tabs.addTab(self.cv_tab, "Control Values")
        if self.yg_font.is_variable_font:
            self.tabs.addTab(self.masters_tab, "Masters")
        self.defaults_pane = defaultsPane(self.yg_font)
        self.tabs.addTab(self.defaults_pane, "Defaults")
        self.layout_obj.addWidget(self.tabs)
        self.setLayout(self.layout_obj)
        self.window().setWindowTitle("Font Info")

    def undo_state_active(self) -> None:
        if not self.yg_font.undo_stack.isActive():
            self.yg_font.undo_stack.setActive(True)

    def closeEvent(self, event) -> None:
        self.hide()

    @pyqtSlot()
    def refresh(self) -> None:
        self.cv_tab.refresh()
        self.defaults_pane.refresh()
        if self.masters_tab:
            self.masters_tab.refresh()

    def event(self, event) -> bool:
        if event.type() == event.Type.WindowActivate:
            self.undo_state_active()
            # self.yg_font.undo_stack.setActive(True)
        return super().event(event)


class mastersWidget(QWidget):
    """A pane for editing masters.

    params:

    owner: The owner of this widget.

    yg_font (ygFont): The font being edited.

    """

    def __init__(self, owner: fontInfoWindow, yg_font: ygFont) -> None:
        super().__init__()
        self.owner = owner
        self.yg_font = yg_font
        self.masters = self.yg_font.masters

        self.layout_obj = QHBoxLayout()
        self.master_list_layout = QVBoxLayout()
        self.button_layout = QHBoxLayout()

        self.master_list = QListWidget()
        self.master_list.addItems(self.masters.names)
        self.current_list_item = self.master_list.item(0)
        self.master_list.setCurrentItem(self.current_list_item)
        self.master_list.itemActivated.connect(self.new_item)

        self._current_master = self.masters.master_by_name(
            self.current_list_item.text()
        )

        # And the edit pane.

        self.edit_pane = masterWidget(
            self.masters, self._current_master[0], self.yg_font
        )

        self.master_list_layout.addWidget(self.master_list)
        add_button = QPushButton("Add")
        del_button = QPushButton("Delete")
        self.button_layout.addWidget(add_button)
        self.button_layout.addWidget(del_button)
        add_button.clicked.connect(self.add_master)
        del_button.clicked.connect(self.del_master)
        self.master_list_layout.addLayout(self.button_layout)
        self.refresh_variants_button = QPushButton("Generate Variant Control Values")
        self.refresh_variants_button.clicked.connect(self.yg_font.refresh_variant_cvs)
        self.master_list_layout.addWidget(self.refresh_variants_button)
        self.layout_obj.addLayout(self.master_list_layout)
        self.layout_obj.addWidget(self.edit_pane)
        self.setLayout(self.layout_obj)

    def current_master_name(self) -> str:
        return self._current_master[1]["name"]

    def current_master_id(self) -> str:
        return self._current_master[0]

    def new_item(self, list_item: QListWidgetItem, forced: bool = False) -> None:
        new_master_name = list_item.text()
        new_master = self.masters.master_by_name(new_master_name)
        if forced or new_master_name != self.current_master_name():
            old_pane = self.layout_obj.itemAt(1)
            self.layout_obj.removeItem(old_pane)
            old_pane.widget().deleteLater()
            self._current_master = new_master
            self.edit_pane = masterWidget(
                self.masters, self.current_master_id(), self.yg_font
            )
            self.layout_obj.addWidget(self.edit_pane)

    @pyqtSlot()
    def add_master(self) -> None:
        master_dict = {}
        axis_tags = self.yg_font.axis_tags
        for a in axis_tags:
            master_dict[a] = 0.0
        master_id = random_id("master")
        master_vals = {"name": master_id, "vals": master_dict}
        self.yg_font.masters.add_master(master_id, master_vals)
        self.refresh()

    @pyqtSlot()
    def del_master(self) -> None:
        self.yg_font.masters.del_by_name(self.current_master_name())
        self.master_list.clear()
        try:
            self._current_master = self.masters.master_by_name(
                self.yg_font.masters.names[0]
            )
        except IndexError:
            return
        self.refresh()

    def refresh(self) -> None:
        if not len(self.yg_font.masters):
            return
        self.master_list.clear()
        self.master_list.addItems(self.yg_font.masters.names)
        matches = self.master_list.findItems(
            self.current_master_name(), Qt.MatchFlag.MatchExactly
        )
        if len(matches) > 0:
            self.master_list.setCurrentItem(matches[0])
        else:
            try:
                current_item = self.master_list.item(0)
                self.master_list.setCurrentItem(current_item)
                self._current_master = self.yg_font.masters.master_by_name(
                    current_item.text()
                )
            except Exception:
                pass
        self.edit_pane.refresh(self._current_master)


class masterWidget(QWidget):
    """A pane for editing a master.

    params:

    masters (ygMasters): The masters for this font.

    m_id (str): The ID of the present master.

    yg_font (ygFont): The font being edited.

    """

    def __init__(self, masters: ygMasters, m_id: str, yg_font: ygFont) -> None:
        super().__init__()
        self.masters = masters
        self.m_id = m_id
        self.master_layout = QVBoxLayout()
        self.name_layout = QHBoxLayout()
        self.name_layout.addWidget(QLabel("Name"))
        self.master_name_widget = masterNameWidget(self.masters, self.m_id)
        self.name_layout.addWidget(self.master_name_widget)
        self.master_layout.addLayout(self.name_layout)
        self.names = []
        axis_tags = yg_font.axis_tags
        for axis_name in axis_tags:
            axis_val_layout = QHBoxLayout()
            axis_val_layout.addWidget(QLabel(axis_name))
            n = masterValWidget(self.masters, self.m_id, axis_name)
            self.names.append(n)
            axis_val_layout.addWidget(n)
            self.master_layout.addLayout(axis_val_layout)
        self.setLayout(self.master_layout)

    def refresh(self, m: Tuple[str, dict]) -> None:
        """Where m is a master tuple (id, dict of axis:val)"""
        self.m_id = m[0]
        self.master_name_widget.refresh(m)
        for n in self.names:
            n.refresh(m)

    # def event(self, event):
    #    print(event)
    #    print(event.type())
    #    print(event.spontaneous())
    #    return super().event(event)


class cvDeltaWidget(QTableView):
    """A table for creating and editing CV Deltas.

    params: cv_source (cvSource): The CV to which deltas will be applied.

    """

    def __init__(self, cv_source: cvSource) -> None:
        super().__init__()
        self.cv_source = cv_source
        self.delta_data = self.cv_source.cvt().get_deltas(
            self.cv_source.current_cv_name()
        )
        self.setModel(self.delta_data)


class cvWidget(QWidget):
    """A pane for making and editing CVs. This class keeps a reference
    to the CV being edited and updates it as the user works with the
    controls.

    params:

    cv_source (cvSource): data from one CV.

    yg_font (ygFont): The font being edited.

    owner: The owner of this widget.

    parent: The parent of this widget.

    delta_pane (bool): Whether a delta pane should be included.

    variant_pane (bool): Whether (if this is a variable font) a
    variant pane should be included.

    """

    def __init__(
        self,
        cv_source: cvSource,
        yg_font: ygFont,
        owner: cvEditPane,
        parent=None,
        delta_pane: bool = True,
        variant_pane: bool = True,
    ):
        super().__init__(parent=parent)
        self.yg_font = yg_font
        self.owner = owner
        self.cv_source = cv_source
        self.layout_obj = QVBoxLayout()

        # Set up tabs

        self.tabs = QTabWidget()
        self.general_tab = QWidget()
        self.link_tab = QWidget()
        self.delta_tab = None
        self.variants_tab = None
        self.masters = None
        if self.yg_font.is_variable_font and variant_pane:
            self.variants_tab = QWidget()
            self.masters = self.yg_font.masters
        self.tabs.addTab(self.general_tab, "General")
        self.tabs.addTab(self.link_tab, "Same As")
        if delta_pane:
            self.delta_tab = QWidget()
            self.tabs.addTab(self.delta_tab, "Deltas")
        if self.variants_tab:
            self.tabs.addTab(self.variants_tab, "Variants")

        # set up general tab

        self.general_tab_layout = QVBoxLayout()
        self.cv_type_widget = cvTypeWidget(self.cv_source)
        #self.cv_color_widget = cvColorWidget(self.cv_source)
        self.cv_axis_widget = cvAxisWidget(self.cv_source)
        self.cv_val_widget = cvValueWidget(self.cv_source)
        self.cv_name_widget = cvNameWidget(self.cv_source, owner=self.owner)
        self.cv_cat_widget = cvUCatWidget(self.cv_source)
        self.cv_suffix_widget = cvSuffixWidget(self.cv_source)

        self.gen_widgets = []

        self.gen_widgets.append(QHBoxLayout())
        self.gen_widgets[-1].addWidget(QLabel("name"))
        self.gen_widgets[-1].addWidget(self.cv_name_widget)

        self.gen_widgets.append(QHBoxLayout())
        self.gen_widgets[-1].addWidget(QLabel("val"))
        self.gen_widgets[-1].addWidget(self.cv_val_widget)

        self.gen_widgets.append(QHBoxLayout())
        self.gen_widgets[-1].addWidget(QLabel("type"))
        self.gen_widgets[-1].addWidget(self.cv_type_widget)

        self.gen_widgets.append(QHBoxLayout())
        self.gen_widgets[-1].addWidget(QLabel("axis"))
        self.gen_widgets[-1].addWidget(self.cv_axis_widget)

        self.gen_widgets.append(QHBoxLayout())
        self.gen_widgets[-1].addWidget(QLabel("cat"))
        self.gen_widgets[-1].addWidget(self.cv_cat_widget)

        self.gen_widgets.append(QHBoxLayout())
        self.gen_widgets[-1].addWidget(QLabel("suffix"))
        self.gen_widgets[-1].addWidget(self.cv_suffix_widget)

        # self.gen_widgets.append(QHBoxLayout())
        # self.gen_widgets[-1].addWidget(QLabel("distance type"))
        # self.gen_widgets[-1].addWidget(self.cv_color_widget)

        # Set up link tab

        self.link_tab_layout = QVBoxLayout()

        self.cv_below_ppem_widget = cvPPEMWidget(self.cv_source, "below")
        self.cv_above_ppem_widget = cvPPEMWidget(self.cv_source, "above")
        self.cv_below_names_widget = cvNamesWidget(
            self.cv_source, "below", self.yg_font, ppem_widget=self.cv_below_ppem_widget
        )
        self.cv_above_names_widget = cvNamesWidget(
            self.cv_source, "above", self.yg_font, ppem_widget=self.cv_above_ppem_widget
        )
        self.cv_below_ppem_widget.name_widget = self.cv_below_names_widget # type: ignore
        self.cv_above_ppem_widget.name_widget = self.cv_above_names_widget # type: ignore

        self.link_widgets = []

        self.link_widgets.append(QHBoxLayout())
        self.link_widgets[-1].addWidget(QLabel("same as"))
        self.link_widgets[-1].addWidget(self.cv_below_names_widget)

        self.link_widgets.append(QHBoxLayout())
        self.link_widgets[-1].addWidget(QLabel("below"))
        self.link_widgets[-1].addWidget(self.cv_below_ppem_widget)
        self.link_widgets[-1].addWidget(QLabel("ppem"))

        self.link_widgets.append(QHBoxLayout())
        self.link_widgets[-1].addWidget(QLabel("and"))

        self.link_widgets.append(QHBoxLayout())
        self.link_widgets[-1].addWidget(QLabel("same as"))
        self.link_widgets[-1].addWidget(self.cv_above_names_widget)

        self.link_widgets.append(QHBoxLayout())
        self.link_widgets[-1].addWidget(QLabel("above"))
        self.link_widgets[-1].addWidget(self.cv_above_ppem_widget)
        self.link_widgets[-1].addWidget(QLabel("ppem"))

        # Set up deltas tab

        self.delta_tab_layout = None
        if self.delta_tab:
            self.delta_tab_layout = QVBoxLayout()
            self.delta_pane = cvDeltaWidget(self.cv_source)
            self.delta_button_layout = QHBoxLayout()
            add_delta_button = QPushButton("Add")
            del_delta_button = QPushButton("Delete")
            add_delta_button.clicked.connect(self.delta_pane.model().new_row) # type: ignore
            del_delta_button.clicked.connect(self.del_delta_row)
            self.delta_button_layout.addWidget(add_delta_button)
            self.delta_button_layout.addWidget(del_delta_button)
            self.delta_tab_layout.addWidget(self.delta_pane)
            self.delta_tab_layout.addLayout(self.delta_button_layout)

        # Set up variants tab

        self.variants_tab_layout = None
        if self.variants_tab:
            self.variants_tab_layout = QVBoxLayout()
            self.var_widgets = []
            self.var_layouts = []
            master_keys = self.masters.keys # type: ignore
            for k in master_keys:
                self.var_layouts.append(QHBoxLayout())
                self.var_layouts[-1].addWidget(QLabel(self.masters.get_master_name(k))) # type: ignore
                self.var_widgets.append(cvVarWidget(k, self.cv_source))
                self.var_layouts[-1].addWidget(self.var_widgets[-1])

        for w in self.gen_widgets:
            self.general_tab_layout.addLayout(w)
        for w in self.link_widgets:
            self.link_tab_layout.addLayout(w)
        if self.variants_tab:
            for w in self.var_layouts:
                self.variants_tab_layout.addLayout(w) # type: ignore

        self.general_tab.setLayout(self.general_tab_layout)
        self.link_tab.setLayout(self.link_tab_layout)
        if self.delta_tab:
            self.delta_tab.setLayout(self.delta_tab_layout) # type: ignore
        if self.variants_tab:
            self.variants_tab.setLayout(self.variants_tab_layout) # type: ignore
        self.layout_obj.addWidget(self.tabs)
        self.setLayout(self.layout_obj)

    @pyqtSlot()
    def del_delta_row(self) -> None:
        i = self.delta_pane.selectedIndexes()
        if len(i) > 0:
            self.delta_pane.model().deleteRows(i[0].row(), 1) # type: ignore

    def refresh(self, cv_source: cvSource) -> None:
        """If we're coming from the source pane, every cv in the
        cvt will have been replaced, so references to it have
        got to be refreshed and the widgets updated as needed.
        Think about whether there's a less awkward way to do
        this.
        """
        self.cv_source = cv_source
        self.cv_name_widget.refresh(self.cv_source)
        self.cv_type_widget.refresh(self.cv_source)
        self.cv_axis_widget.refresh(self.cv_source)
        #self.cv_color_widget.refresh(self.cv_source)
        self.cv_val_widget.refresh(self.cv_source)
        self.cv_cat_widget.refresh(self.cv_source)
        self.cv_suffix_widget.refresh(self.cv_source)
        self.cv_below_ppem_widget.refresh(self.cv_source)
        self.cv_above_ppem_widget.refresh(self.cv_source)
        self.cv_below_names_widget.refresh(self.cv_source)
        self.cv_above_names_widget.refresh(self.cv_source)
        if self.masters:
            for w in self.var_widgets:
                w.refresh(self.cv_source)

    def fixup(self) -> None:
        self.cv_name_widget.fixup()
        self.cv_type_widget.fixup()
        self.cv_axis_widget.fixup()
        #self.cv_color_widget.fixup()
        self.cv_val_widget.fixup()
        self.cv_cat_widget.fixup()
        self.cv_suffix_widget.fixup()
        self.cv_below_ppem_widget.fixup()
        self.cv_above_ppem_widget.fixup()
        self.cv_below_names_widget.fixup()
        self.cv_above_names_widget.fixup()
        if self.masters:
            for w in self.var_widgets:
                w.fixup()

    # def event(self, event):
    #    print(event)
    #    return super().event(event)


class defaultsPane(QWidget):
    """A tabbed pane holding two panes for editing font-wide
    defaults.

    params:

    yg_font (ygFont): The font being edited.

    """

    def __init__(self, yg_font: ygFont) -> None:
        super().__init__()
        self.layout_obj = QVBoxLayout()
        self.tabs = QTabWidget()
        self.round_widget = hintRoundWidget(yg_font)
        self.misc_widget = miscDefaultsWidget(yg_font)
        self.tabs.addTab(self.round_widget, "Rounding")
        self.tabs.addTab(self.misc_widget, "Miscellaneous")
        self.layout_obj.addWidget(self.tabs)
        self.setLayout(self.layout_obj)

    def refresh(self) -> None:
        self.round_widget.refresh()
        self.misc_widget.refresh()


class hintRoundWidget(QWidget):
    """Widget for editing the initial round state of the seven types of
    hint.

    params:

    yg_font (ygFont): The font being edited.

    """

    def __init__(self, yg_font: ygFont) -> None:
        super().__init__()
        self.yg_font = yg_font
        self.defaults = yg_font.defaults
        self.layout_obj = QVBoxLayout()
        self.ignore_signal = False

        self.anchor_layout = QHBoxLayout()
        self.anchor_checkbox = QCheckBox("Anchor")
        self.anchor_layout.addWidget(self.anchor_checkbox)
        self.layout_obj.addLayout(self.anchor_layout)

        self.stem_layout = QHBoxLayout()
        self.stem_checkbox = QCheckBox("Stem")
        self.stem_layout.addWidget(self.stem_checkbox)
        self.layout_obj.addLayout(self.stem_layout)

        # self.blackdist_layout = QHBoxLayout()
        # self.blackdist_checkbox = QCheckBox("Black distance")
        # self.blackdist_layout.addWidget(self.blackdist_checkbox)
        # self.layout_obj.addLayout(self.blackdist_layout)

        # self.whitedist_layout = QHBoxLayout()
        # self.whitedist_checkbox = QCheckBox("White distance")
        # self.whitedist_layout.addWidget(self.whitedist_checkbox)
        # self.layout_obj.addLayout(self.whitedist_layout)

        # self.graydist_layout = QHBoxLayout()
        # self.graydist_checkbox = QCheckBox("Gray distance")
        # self.graydist_layout.addWidget(self.graydist_checkbox)
        # self.layout_obj.addLayout(self.graydist_layout)

        self.shift_layout = QHBoxLayout()
        self.shift_checkbox = QCheckBox("Shift")
        self.shift_layout.addWidget(self.shift_checkbox)
        self.layout_obj.addLayout(self.shift_layout)

        self.align_layout = QHBoxLayout()
        self.align_checkbox = QCheckBox("Align")
        self.align_layout.addWidget(self.align_checkbox)
        self.layout_obj.addLayout(self.align_layout)

        self.interpolate_layout = QHBoxLayout()
        self.interpolate_checkbox = QCheckBox("Interpolate")
        self.interpolate_layout.addWidget(self.interpolate_checkbox)
        self.layout_obj.addLayout(self.interpolate_layout)

        self.refresh()

        self.anchor_checkbox.stateChanged.connect(self.button_state_changed)
        self.stem_checkbox.stateChanged.connect(self.button_state_changed)
        # self.blackdist_checkbox.stateChanged.connect(self.button_state_changed)
        # self.whitedist_checkbox.stateChanged.connect(self.button_state_changed)
        # self.graydist_checkbox.stateChanged.connect(self.button_state_changed)
        self.shift_checkbox.stateChanged.connect(self.button_state_changed)
        self.align_checkbox.stateChanged.connect(self.button_state_changed)
        self.interpolate_checkbox.stateChanged.connect(self.button_state_changed)

        self.setLayout(self.layout_obj)

    @pyqtSlot()
    def button_state_changed(self) -> None:
        if self.ignore_signal:
            return
        r = {}
        r["anchor"] = self.anchor_checkbox.isChecked()
        r["stem"] = self.stem_checkbox.isChecked()
        # r["blackdist"] = self.blackdist_checkbox.isChecked()
        # r["whitedist"] = self.whitedist_checkbox.isChecked()
        # r["graydist"] = self.graydist_checkbox.isChecked()
        r["shift"] = self.shift_checkbox.isChecked()
        r["align"] = self.align_checkbox.isChecked()
        r["interpolate"] = self.interpolate_checkbox.isChecked()
        self.defaults.set_rounding_defaults(r)

    def fixup(self) -> None:
        pass

    def refresh(self) -> None:
        self.ignore_signal = True
        self.anchor_checkbox.setChecked(self.defaults.rounding_state("anchor"))
        self.stem_checkbox.setChecked(self.defaults.rounding_state("stem"))
        # self.blackdist_checkbox.setChecked(self.defaults.rounding_state("blackdist"))
        # self.whitedist_checkbox.setChecked(self.defaults.rounding_state("whitedist"))
        # self.graydist_checkbox.setChecked(self.defaults.rounding_state("graydist"))
        self.shift_checkbox.setChecked(self.defaults.rounding_state("shift"))
        self.align_checkbox.setChecked(self.defaults.rounding_state("align"))
        self.interpolate_checkbox.setChecked(
            self.defaults.rounding_state("interpolate")
        )
        self.ignore_signal = False


class miscDefaultsWidget(QWidget):
    """GUI for setting defaults that have nothing to do with rounding.

    params:

    yg_font (ygFont): the font now being edited.
    """

    def __init__(self, yg_font: ygFont) -> None:
        super().__init__()
        self.yg_font = yg_font
        self.defaults = yg_font.defaults
        self.layout_obj = QGridLayout()
        self.layout_obj.setHorizontalSpacing(20)
        self.layout_obj.setVerticalSpacing(0)
        self.ignore_signal = False

        self.tt_defaults = QCheckBox("Use TrueType defaults")
        s =  "<span>If</span> this is checked, all instructions in the CVT program that set "
        s += "defaults will be ignored."
        self.tt_defaults.setToolTip(s)
        self.tt_defaults.stateChanged.connect(self.toggle_tt_defaults)

        self.init_graphics = QCheckBox("Initialize graphics")
        s =  "<span>If</span> this is checked, code to initialize graphics variables will "
        s += "be inserted at the beginning of each glyph program. This should "
        s += "normally be left unchecked, as these variables are seldom "
        s += "consulted in modern TrueType hinting."
        self.init_graphics.setToolTip(s)
        self.init_graphics.stateChanged.connect(self.toggle_init_graphics)

        self.assume_always_y = QCheckBox("Assume axis always y")
        s =  "<span>Instructs</span> the compiler to assume that you are only hinting on "
        s += "y axis. When this is checked, Ygt will perform several optimizations."
        self.assume_always_y.setToolTip(s)
        self.assume_always_y.stateChanged.connect(self.toggle_assume_always_y)

        self.counterclockwise = QCheckBox("Outer contours counter-clockwise")
        s =  "<span>When</span> outer contours are counter-clockwise, Ygt will guess distance types "
        s += "wrongly. Check this box and it will guess correctly. You should also "
        s += "select View→Point coordinates since contours will probably be reversed "
        s += "when you generate the font, throwing off the point indices."
        self.counterclockwise.setToolTip(s)
        self.counterclockwise.stateChanged.connect(self.toggle_counterclockwise)

        self.cleartype = QCheckBox("Cleartype")
        s =  "<span>If</span> this box is checked, MS Windows will render your font"
        s += "in native ClearType mode."
        self.cleartype.setToolTip(s)
        self.cleartype.stateChanged.connect(self.toggle_cleartype)

        self.mergemode = QCheckBox("Merge mode")
        s =  "<span>Check</span> this box if you are hinting a font containing hints that you "
        s += "want to retain. Glyphs that you’ve hinted will use your hints. Your functions "
        s += "and control values will be appended to the ones already in the font."
        self.mergemode.setToolTip(s)
        self.mergemode.stateChanged.connect(self.toggle_mergemode)

        self.replaceprep = QCheckBox("Replace CVT program")
        s =  "<span>Only</span> available when merge mode is active. If you check this box, "
        s += "the CVT program generated by Ygt will replace the one in the font. Otherwise, "
        s += "the Ygt CVT program will be appended to the existing one."
        self.replaceprep.setToolTip(s)
        self.replaceprep.stateChanged.connect(self.toggle_replaceprep)

        self.functionbase = functionBaseWidget(self.defaults)
        s =  "In merge mode, <span>set</span> 'function-base' to a non-zero value if Ygt guesses wrongly "
        s += "about the highest-numbered function in the font to which you are adding hints."
        self.functionbase.setToolTip(s)

        self.layout_obj.addWidget(self.tt_defaults, 1, 1)
        self.layout_obj.addWidget(self.init_graphics, 2, 1)
        self.layout_obj.addWidget(self.cleartype, 3, 1)
        self.layout_obj.addWidget(self.assume_always_y, 4, 1)
        self.layout_obj.addWidget(self.counterclockwise, 1, 2)
        self.layout_obj.addWidget(self.mergemode, 2, 2)
        self.layout_obj.addWidget(self.replaceprep, 3, 2)
        self.functionbase.setFixedWidth(int(self.functionbase.width() /4))
        function_base_layout = QHBoxLayout()
        function_base_layout.addWidget(QLabel("Function base"))
        function_base_layout.addWidget(self.functionbase, alignment = Qt.AlignmentFlag.AlignLeft)
        self.layout_obj.addLayout(function_base_layout, 4, 2)

        self.setLayout(self.layout_obj)

        self.refresh()

    def toggle_counterclockwise(self) -> None:
        if self.ignore_signal:
            return
        if self.counterclockwise.isChecked():
            self.defaults.set_default({"counterclockwise": True})
        else:
            self.defaults.del_default("counterclockwise")

    @pyqtSlot()
    def toggle_tt_defaults(self) -> None:
        if self.ignore_signal:
            return
        if self.tt_defaults.isChecked():
            self.defaults.set_default({"use-truetype-defaults": True})
        else:
            self.defaults.del_default("use-truetype-defaults")

    @pyqtSlot()
    def toggle_init_graphics(self) -> None:
        if self.ignore_signal:
            return
        if not self.init_graphics.isChecked():
            self.defaults.set_default({"init-graphics": False})
        else:
            self.defaults.del_default("init-graphics")

    @pyqtSlot()
    def toggle_assume_always_y(self) -> None:
        if self.ignore_signal:
            return
        if self.assume_always_y.isChecked():
            self.defaults.set_default({"assume-always-y": True})
        else:
            self.defaults.del_default("assume-always-y")

    @pyqtSlot()
    def toggle_cleartype(self) -> None:
        if self.ignore_signal:
            return
        if self.cleartype.isChecked():
            self.defaults.set_default({"cleartype": True})
        else:
            self.defaults.del_default("cleartype")

    @pyqtSlot()
    def toggle_mergemode(self) -> None:
        if self.ignore_signal:
            return
        if self.mergemode.isChecked():
            self.defaults.set_default({"merge-mode": True})
            self.replaceprep.setEnabled(True)
            self.functionbase.setEnabled(True)
        else:
            self.defaults.del_default("merge-mode")
            self.replaceprep.setEnabled(False)
            self.functionbase.setEnabled(False)

    @pyqtSlot()
    def toggle_replaceprep(self) -> None:
        if self.ignore_signal:
            return
        if bool(self.defaults.get_default("merge-mode")) and self.replaceprep.isChecked():
            self.defaults.set_default({"replace-prep": True})
        else:
            self.defaults.del_default("replace-prep")

    def fixup(self) -> None:
        pass

    def refresh(self, ign: bool = True) -> None:
        self.ignore_signal = ign

        t = self.defaults.get_default("use-truetype-defaults")
        self.tt_defaults.setChecked(bool(t))

        t = self.defaults.get_default("init-graphics")
        if t == None:
            t = True
        self.init_graphics.setChecked(t)

        t = self.defaults.get_default("assume-always-y")
        self.assume_always_y.setChecked(bool(t))

        t = self.defaults.get_default("cleartype")
        self.cleartype.setChecked(bool(t))

        t = self.defaults.get_default("merge-mode")
        self.mergemode.setChecked(bool(t))

        t = self.defaults.get_default("replace-prep")
        self.replaceprep.setChecked(bool(t))

        if not bool(t):
            self.replaceprep.setEnabled(False)
            self.functionbase.setEnabled(False)

        self.ignore_signal = False


class makeCVDialog(QDialog, cvSource):
    """A dialog for creating a cv. This doesn't edit the cvt source
    directly, but instead works on a fragment of cv code to be
    added when the accept() function is called.

    params:

    p1 (ygPoint): The first selected point.

    p2 (ygPoint): The second selected point; or None if only one
    point is selected.

    yg_glyph (ygGlyph): The current glyph.

    preferences (ygPreferences): the preferences for this app.

    """
    def __init__(self, p1: ygPoint, p2: ygPoint, yg_glyph: ygGlyph, preferences: ygPreferences) -> None:
        super().__init__()
        self.top_window = preferences.top_window()
        self.yg_font = yg_glyph.yg_font
        self._cvt = self.yg_font.cvt
        self.cv = {}
        self.cv_name = ""
        self.axis = preferences.top_window().current_axis
        self.cv["axis"] = self.axis
        if p2 != None:
            init_type = "dist"
            origin_indices = [p1.index, p2.index]
        else:
            init_type = "pos"
            origin_indices = [p1.index]
        self.cv["type"] = init_type
        val = 0
        if self.axis == "y":
            if init_type == "pos":
                val = p1.font_y
            else:
                val = abs(p1.font_y - p2.font_y)
        else:
            if init_type == "pos":
                val = p1.font_x
            else:
                val = abs(p1.font_x - p2.font_x)
        self.cv["val"] = val
        self.cv["origin"] = {"glyph": yg_glyph.gname, "ptnum": origin_indices}

        self.layout_obj = QVBoxLayout()

        self.pane = cvWidget(
            self, self.yg_font, None, delta_pane=False, variant_pane=False
        )

        # Set up buttons

        QBtn = (
            QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        )
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

        self.layout_obj.addWidget(self.pane)
        self.layout_obj.addWidget(self.buttonBox)
        self.setLayout(self.layout_obj)
        self.setWindowTitle("Make Control Value")

    def send_error_message(self, d: dict) -> None:
        self.yg_font.send_error_message(d)

    def cvt(self) -> ygcvt:
        return self._cvt

    def current_cv(self) -> dict:
        return self.cv

    def current_cv_name(self) -> str:
        return self.cv_name

    def set_cv_name(self, s: str):
        self.cv_name = s

    def from_current_cv(self, s: str) -> None:
        try:
            return self.cv[s]
        except KeyError:
            pass

    def set_in_current_cv(self, k: str, s: Any, fallback=None) -> None:
        if s == "None" or s == "" or s == None:
            if fallback != None:
                self.cv[k] = fallback
            else:
                if k in self.cv:
                    del self.cv[k]
                #if k in self._current_cv:
                #    del self._current_cv[k]
        else:
            self.cv[k] = s

    def has_key(self, k: str) -> bool:
        return k in self.cv

    def del_key(self, k: str) -> None:
        try:
            del self.cv[k]
        except KeyError:
            pass

    def accept(self) -> None:
        self.pane.fixup()
        if self.cv_name and len(self.cv) > 0:
            self._cvt.add_cv(self.cv_name, self.cv)
        super().accept()

    def showEvent(self, event) -> None:
        self.yg_font.undo_stack.setActive(True)


class cvNameWidget(QLineEdit):
    """A widget for editing the name of a cv. Disable when it shouldn't
    be edited.

    params:

    cv_source (cvSource): Data for currently selected CV.

    owner: The owner of this widget.
    """

    def __init__(self, cv_source: cvSource, owner: Any = None) -> None:
        super().__init__()
        self.owner = owner
        self.cv_source = cv_source
        self.cv_name = self.cv_source.current_cv_name()
        self.dirty = False
        if len(self.cv_name) > 0:
            self.setText(self.cv_name)
            if self.cv_name != NEW_CV_NAME:
                self.setEnabled(False)
        if self.isEnabled():
            self.editingFinished.connect(self.text_changed)
            self.textChanged.connect(self.set_dirty)
        self.last_val = self.cv_name

    def _text(self) -> str:
        return self.text().strip()

    @pyqtSlot()
    def set_dirty(self) -> None:
        self.dirty = True

    def set_clean(self) -> None:
        self.dirty = False

    def fixup(self) -> None:
        t = self._text()
        if self.isEnabled() and self.dirty and t != self.last_val:
            # If the cv name changes, the original cv in the source tree
            # has to be deleted and a new cv entered under the new name.
            # We can either do that (at some risk of bugginess) or disable
            # the cv name widget when the cv is being edited (as opposed
            # to created).
            old_name = self.cv_source.current_cv_name()
            self.cv_source.set_cv_name(t)
            if self.owner != None:
                self.owner.change_name_in_list(t)
                self.owner.rename_cv(old_name, t)
            self.last_val = t
            self.set_clean()

    @pyqtSlot()
    def text_changed(self) -> None:
        self.fixup()

    def refresh(self, cv_source: cvSource) -> None:
        self.cv_source = cv_source
        self.setText(self.cv_source.current_cv_name())
        self.set_clean()


class cvTypeWidget(QComboBox):
    """Widget for choosing a CV type.

    params:

    cv_source (cvSource): Data for currently selected CV.

    """

    def __init__(self, cv_source: cvSource) -> None:
        super().__init__()
        self.cv_source = cv_source
        self.addItem("pos")
        self.addItem("dist")
        t = self.cv_source.from_current_cv("type")
        self.setCurrentText((lambda: "y" if not t else t)())
        self.currentTextChanged.connect(self.text_changed)
        self.last_val = self.currentText()

    def _text(self) -> str:
        return self.currentText()

    def fixup(self) -> None:
        t = self.currentText()
        if t != self.last_val:
            self.cv_source.set_in_current_cv("type", self.currentText(), None)
            self.last_val = t

    @pyqtSlot()
    def text_changed(self) -> None:
        self.fixup()

    def refresh(self, cv_source: cvSource) -> None:
        self.cv_source = cv_source
        self.setCurrentText(self.cv_source.from_current_cv("type"))


class cvColorWidget(QComboBox):
    """Widget for choosing a distance type.

    params:

    cv_source (cvSource): Data for currently selected CV.

    """

    def __init__(self, cv_source: cvSource) -> None:
        super().__init__()
        self.cv_source = cv_source
        self.addItem("None")
        self.addItem("black")
        self.addItem("white")
        self.addItem("gray")
        col = self.cv_source.from_current_cv("col")
        self.setCurrentText((lambda: "None" if not col else col)())
        self.currentTextChanged.connect(self.text_changed)
        self.last_val = self.currentText()

    def _text(self) -> str:
        return self.currentText()

    def fixup(self) -> None:
        new_text = self.currentText()
        if new_text != self.last_val:
            if self.cv_source.current_cv():
                if new_text != "None":
                    self.cv_source.set_in_current_cv("col", new_text, None)
                else:
                    self.cv_source.del_key("col")
            self.last_val = new_text

    @pyqtSlot()
    def text_changed(self) -> None:
        self.fixup()

    def refresh(self, cv_source: cvSource) -> None:
        self.cv_source = cv_source
        if self.cv_source.has_key("col"):
            self.setCurrentText(self.cv_source.from_current_cv("col"))
        else:
            self.setCurrentText("None")


class cvAxisWidget(QComboBox):
    """Widget for choosing an axis for a CV.

    params:

    cv_source (cvSource): Data for currently selected CV.

    """

    def __init__(self, cv_source: cvSource) -> None:
        super().__init__()
        self.cv_source = cv_source
        self.addItem("y")
        self.addItem("x")
        axis = self.cv_source.from_current_cv("axis")
        self.setCurrentText((lambda: "y" if not axis else axis)())
        self.currentTextChanged.connect(self.text_changed)
        self.last_val = self.currentText()

    def _text(self) -> str:
        return self.currentText()

    def fixup(self) -> None:
        t = self.currentText()
        if t != self.last_val:
            self.cv_source.set_in_current_cv("axis", self.currentText(), None)
            self.last_val = t

    @pyqtSlot()
    def text_changed(self) -> None:
        self.fixup()

    def refresh(self, cv_source) -> None:
        self.cv_source = cv_source
        if self.cv_source.has_key("axis"):
            self.setCurrentText(self.cv_source.from_current_cv("axis"))
        else:
            self.setCurrentText("y")


class cvUCatWidget(QComboBox):
    """Widget for choosing a category for a CV.

    params:

    cv_source (cvSource): Data for currently selected CV.

    """

    def __init__(self, cv_source: cvSource) -> None:
        super().__init__()
        self.cv_source = cv_source
        self.cats = unicode_cat_names.values()
        self.addItem("None")
        for c in self.cats:
            self.addItem(c)
        ct = "None"
        if self.cv_source.has_key("cat"):
            ct = unicode_cat_names[self.cv_source.from_current_cv("cat")]
        self.setCurrentText(ct)
        self.currentTextChanged.connect(self.text_changed)
        self.last_val = self.currentText()

    def _text(self) -> str:
        return self.currentText()

    def fixup(self) -> None:
        new_text = self.currentText()
        if new_text != self.last_val:
            if not new_text or new_text == "None":
                self.cv_source.del_key("cat")
            else:
                self.cv_source.set_in_current_cv(
                    "cat", reverse_unicode_cat_names[new_text], None
                )
            self.last_val = new_text

    @pyqtSlot()
    def text_changed(self) -> None:
        self.fixup()

    def refresh(self, cv_source) -> None:
        self.cv_source = cv_source
        if self.cv_source.has_key("cat"):
            self.setCurrentText(
                unicode_cat_names[self.cv_source.from_current_cv("cat")]
            )
        else:
            self.setCurrentText("None")


class cvSuffixWidget(QLineEdit):
    """Widget for specifying a suffix (the CV is only available for glyphs with
    this suffix).

    params:

    cv_source (cvSource): Data for currently selected CV.

    """

    def __init__(self, cv_source: cvSource) -> None:
        super().__init__()
        self.cv_source = cv_source
        self.dirty = False
        suff = self.cv_source.from_current_cv("suffix")
        self.setText((lambda: "None" if not suff else suff)())
        self.editingFinished.connect(self.text_changed)
        self.textChanged.connect(self.set_dirty)
        self.last_val = self.text()

    def _text(self) -> str:
        return self.text().strip()

    @pyqtSlot()
    def set_dirty(self) -> None:
        self.dirty = True

    def set_clean(self) -> None:
        self.dirty = False

    def fixup(self) -> None:
        if self.dirty:
            self.cv_source.set_in_current_cv("suffix", self._text(), None)
            self.set_clean()

    @pyqtSlot()
    def text_changed(self) -> None:
        t = self._text()
        if t != self.last_val:
            self.fixup()
            self.last_val = t

    def refresh(self, cv_source: cvSource) -> None:
        self.cv_source = cv_source
        suff = self.cv_source.from_current_cv("suffix")
        self.setText((lambda: "None" if not suff else suff)())
        self.set_clean()


class cvVarWidget(QLineEdit):
    """Widget for editing variant CVs for the cvar table.

    params:

    var_id (str): ID of the master associated with this widget.

    cv_source (cvSource): Data for currently selected CV.

    """

    def __init__(self, var_id: str, cv_source: cvSource) -> None:
        super().__init__()
        self.var_id = var_id
        self.cv_source = cv_source
        self.dirty = False
        v = self.cv_source.from_current_cv("var")
        k = "None"
        if v and self.var_id in v:
            k = v[self.var_id]
        self.setText(str(k))
        self.setValidator(QIntValidator(-9999, 9999))
        self.editingFinished.connect(self.text_changed)
        self.textChanged.connect(self.set_dirty)
        self.last_val = str(k)

    def _text(self) -> str:
        return self.text().strip()

    @pyqtSlot()
    def set_dirty(self) -> None:
        self.dirty = True

    def set_clean(self) -> None:
        self.dirty = False

    def fixup(self) -> None:
        """The behavior we want is this: if the widget is blank,
        delete the value for this master and display "None"
        in the widget. If widget has a valid integer, plug
        that in.
        """
        if self.dirty:
            new_text = self._text()
            current_vars = self.cv_source.from_current_cv("var")
            if not current_vars:
                current_vars = {}
            if new_text:
                current_vars[self.var_id] = int(new_text)
                self.cv_source.set_in_current_cv("var", current_vars, None)
            else:
                if len(current_vars) == 0:
                    self.cv_source.del_key("var")
                    self.refresh(self.cv_source)
            self.set_clean()

    @pyqtSlot()
    def text_changed(self) -> None:
        t = self._text()
        if t != self.last_val:
            self.fixup()
            self.last_val = t

    def refresh(self, cv_source: cvSource) -> None:
        self.cv_source = cv_source
        v = self.cv_source.from_current_cv("var")
        k = "None"
        if v and self.var_id in v:
            k = v[self.var_id]
        self.setText(str(k))
        self.set_clean()


class functionBaseWidget(QLineEdit):
    def __init__(self, defaults):
        super().__init__()
        self.defaults = defaults
        init_val = self.defaults.get_default("function-base")
        if init_val == None:
            init_val = 0
        self.setText(str(init_val))
        self.setValidator(QIntValidator(0, 255))
        self.editingFinished.connect(self.text_changed)
        self.last_val = self.text()

    def _text(self) -> str:
        return self.text().strip()

    def fixup(self) -> None:
        t = self._text()
        if t != self.last_val:
            try:
                i = int(t)
            except ValueError:
                i = 0
            self.defaults.set_default({"function-base": i})

    @pyqtSlot()
    def text_changed(self) -> None:
        t = self._text()
        if t != self.last_val:
            self.fixup()
            self.last_val = t

    def refresh(self) -> None:
        fb = self.defaults.get_default("function-base")
        if fb == None:
            self.setText("0")
        else:
            self.setText(str(fb))


class cvValueWidget(QLineEdit):
    """Widget for editing the value of a CV.

    params:

    cv_source (cvSource): Data for currently selected CV.

    """

    def __init__(self, cv_source: cvSource) -> None:
        super().__init__()
        self.cv_source = cv_source
        self.dirty = False
        self.setText(str(self.cv_source.from_current_cv("val")))
        self.setValidator(QIntValidator(-9999, 9999))
        self.editingFinished.connect(self.text_changed)
        self.textChanged.connect(self.set_dirty)
        self.last_val = self.text()

    def _text(self) -> str:
        return self.text().strip()

    def set_dirty(self) -> None:
        self.dirty = True

    def set_clean(self) -> None:
        self.dirty = False

    def fixup(self) -> None:
        t = self._text()
        if t != self.last_val:
            try:
                i = int(t)
            except ValueError:
                i = 0
            self.cv_source.set_in_current_cv("val", i, fallback = 0)
            self.set_clean()

    @pyqtSlot()
    def text_changed(self) -> None:
        t = self._text()
        if t != self.last_val:
            self.fixup()
            self.last_val = t

    def refresh(self, cv_source: cvSource) -> None:
        self.cv_source = cv_source
        i = self.cv_source.from_current_cv("val")
        self.setText((lambda: "0" if not i else str(i))())
        self.set_clean()


class cvPPEMWidget(QLineEdit):
    """Widget for editing a "ppem" value in the "same as" pane.

    params:

    cv_source (cvSource): Data for currently selected CV.

    above_below (str): Indicating which widget pair is being edited.

    """

    def __init__(self, cv_source: cvSource, above_below: str) -> None:
        super().__init__()
        self.cv_source = cv_source
        self.above_below = above_below
        self.name_widget: Optional[cvNameWidget] = None
        self.dirty = False
        s = self.cv_source.from_current_cv("same-as")
        v = "40"
        if s:
            try:
                self.setText(s[self.above_below])
            except Exception:
                self.setText("40")
        self.setValidator(QIntValidator(0, 500))
        self.editingFinished.connect(self.text_changed)
        self.textChanged.connect(self.set_dirty)
        self.last_val = "40"

    def _text(self) -> str:
        return self.text().strip()

    @pyqtSlot()
    def set_dirty(self) -> None:
        self.dirty = True

    def set_clean(self) -> None:
        self.dirty = False

    def fixup(self) -> None:
        new_name = self.name_widget._text()
        name_valid = new_name != "None"
        try:
            new_val = int(self._text())
        except Exception:
            new_val = 40
            self.setText(str(new_val))
        if name_valid:
            s = self.cv_source.from_current_cv("same-as")
            if not s:
                s = {self.above_below: {}}
            if not self.above_below in s:
                s[self.above_below] = {}
            s[self.above_below]["cv"] = new_name
            s[self.above_below]["ppem"] = new_val
            self.cv_source.set_in_current_cv("same-as", s, None)
        else:
            if self.cv_source.has_key("same-as"):
                s = self.cv_source.from_current_cv("same-as")
                if self.above_below in s:
                    del s[self.above_below]
                if len(s) == 0:
                    self.cv_source.del_key("same-as")
        self.set_clean()

    @pyqtSlot()
    def text_changed(self) -> None:
        t = self._text()
        if t != self.last_val:
            self.fixup()
            self.last_val = t

    def refresh(self, cv_source: cvSource) -> None:
        self.cv_source = cv_source
        p = "40"
        s = self.cv_source.from_current_cv("same-as")
        if s and self.above_below in s:
            p = str(s[self.above_below]["ppem"])
        self.setText(p)
        self.set_clean()


class cvNamesWidget(QComboBox):
    """Widget for choosing a CV name in the "same as" pane.

    params:

    cv_source (cvSource): Data for currently selected CV.

    above_below (str): Indicating which widget pair is being edited.

    yg_font (ygFont): The font being edited.

    ppem_widget: The associated cvPPEMWidget (if any).

    """

    def __init__(
            self,
            cv_source: cvSource,
            above_below: str,
            yg_font: ygFont,
            ppem_widget: cvPPEMWidget = None
        ) -> None:
        super().__init__()
        self.cv_source = cv_source
        self.yg_font = yg_font
        self.cvt = yg_font.cvt
        self.above_below = above_below
        self.ppem_widget = ppem_widget
        cv_list = self.cvt.get_list(None)
        self.addItem("None")
        for c in cv_list:
            self.addItem(c)
        n = "None"
        cv = self.cv_source.from_current_cv("same-as")
        if cv and self.above_below in cv:
            n = cv[self.above_below]["cv"]
        self.setCurrentText(n)
        self.currentTextChanged.connect(self.text_changed)

    def _text(self) -> str:
        return self.currentText()

    def fixup(self) -> None:
        if self.ppem_widget:
            self.ppem_widget.fixup()

    @pyqtSlot()
    def text_changed(self) -> None:
        self.fixup()

    def refresh(self, cv_source: cvSource) -> None:
        self.cv_source = cv_source
        n = "None"
        if self.cv_source.has_key("same-as"):
            s = self.cv_source.from_current_cv("same-as")
            if self.above_below in s:
                n = s[self.above_below]["cv"]
        self.setCurrentText(n)


class masterNameWidget(QLineEdit):
    """Widget for editing the name of a master. This name is for
    display only: the master is referenced by an immutable id.

    params:

    masters (ygMasters): The collection of masters for this font.

    m_id (str): The ID of the master associated with this widget.

    """

    def __init__(self, masters: ygMasters, m_id: str) -> None:
        super().__init__()
        self.masters = masters
        self.m_id = m_id
        master_name = self.masters.get_master_name(self.m_id)
        self.setText(master_name)
        self.dirty = False
        self.editingFinished.connect(self.text_changed)
        self.textChanged.connect(self.set_dirty)
        self.last_val = master_name

    def _text(self) -> str:
        return self.text().strip()

    @pyqtSlot()
    def set_dirty(self) -> None:
        self.dirty = True

    def set_clean(self) -> None:
        self.dirty = False

    def refresh(self, m: tuple) -> None:
        """refresh is for updating editing widgets from the model."""
        self.m_id = m[0]
        self.setText(self.masters.get_master_name(self.m_id))
        self.set_clean()

    def fixup(self) -> None:
        """fixup is for changing the model based on what's in the editing
        widgets.
        """
        self.masters.set_master_name(self.m_id, self._text())
        self.set_clean()

    @pyqtSlot()
    def text_changed(self) -> None:
        t = self._text()
        if self.dirty and t != self.last_val:
            self.fixup()
            self.last_val = t


class masterValWidget(QLineEdit):
    """Widget for editing the value of a master (-1.0 to 1.0)

    params:

    masters (ygMasters): The collection of masters for this font.

    m_id (str): The ID of the master associated with this widget.

    axis: The variation axis associated with this widget.

    """

    def __init__(self, masters: ygMasters, m_id: str, axis: str) -> None:
        super().__init__()
        self.masters = masters
        self.m_id = m_id
        self.axis = axis
        self.init_val = self.masters.get_axis_value(self.m_id, self.axis)
        if self.init_val == None:
            self.init_val = 0.0
        self.setText(str(self.init_val))
        self.setValidator(QDoubleValidator(-1.0, 1.0, 4))
        self.dirty = False
        self.editingFinished.connect(self.text_changed)
        self.textChanged.connect(self.set_dirty)
        self.last_val = str(self.init_val)

    def _text(self) -> str:
        return self.text().strip()

    @pyqtSlot()
    def set_dirty(self) -> None:
        self.dirty = True

    def set_clean(self) -> None:
        self.dirty = False

    def refresh(self, m: tuple) -> None:
        """refresh is for updating editing widgets from the model."""
        if m:
            self.m_id = m[0]
        self.setText(str(self.masters.get_axis_value(self.m_id, self.axis)))
        self.set_clean()

    def fixup(self) -> None:
        """fixup is for changing the model based on what's in the editing
        widgets.
        """
        try:
            v = self._text()
            if v == "" or v == "0.0":
                self.masters.del_axis(self.m_id, self.axis)
            else:
                try:
                    self.masters.set_axis_value(self.m_id, self.axis, float(v))
                except Exception:
                    self.masters.set_axis_value(self.m_id, self.axis, 0.0)
                self.refresh(None)
            self.set_clean()
        except Exception as e:
            print(e)
            pass

    @pyqtSlot()
    def text_changed(self) -> None:
        t = self._text()
        if self.dirty and t != self.last_val:
            self.fixup()
            self.last_val = t



from lxml import etree

"""
YAnchor
ResYAnchor
[YNoRound, YAnchor]
[YNoRound, ResYAnchor]
ResYAnchor
YInterpolate
YIPAnchor
YDist
ResYDist
[YNoRound, YDist]
[YNoRound, ResYDist]
YLink
ResYLink
[YNoRound, YLink]
[YNoRound, ResYLink]
YShift
smooth
"""

def build_input_output(self):
    pass

def build_defaults(self):
    pass

def build_cvt_settings(self):
    pass

def build_cvt(self):
    pass

def build_cvar_from_masters(self):
    pass

def build_functions(self):
    pass

def build_macros(self):
    pass

def build_glyph_program(g, y_doc[k][g], xgf_doc):
    pass

def ygridfit_parse_obj(y_doc, single_glyph=None, glyph_list=None):
    y_keys = y_doc.keys()


    for k in y_keys:
        if k == "font":
            build_input_output(y_doc[k], xgf_doc)
        elif k == "defaults":
            build_defaults(y_doc[k], xgf_doc)
        elif k == "prep":
            build_cvt_settings(y_doc[k], y_doc["cvt"], xgf_doc)
        elif k == "cvt":
            build_cvt(y_doc[k], xgf_doc)
        elif k == "masters":
            try:
                build_cvar_from_masters(y_doc, xgf_doc)
            except Exception as e:
                pass
        #elif k == "cvar":
        #    if not "masters" in y_doc:
        #        build_cvar(y_doc[k], xgf_doc)
        elif k == "functions":
            build_functions(y_doc[k], xgf_doc)
        elif k == "macros":
            build_macros(y_doc[k], xgf_doc)
        elif k == "glyphs":
            if not single_glyph and not glyph_list:
                g_keys = y_doc[k].keys()
            elif glyph_list and not single_glyph:
                g_keys = glyph_list
            else:
                g_keys = [single_glyph]
            for g in g_keys:
                try:
                    build_glyph_program(g, y_doc[k][g], xgf_doc)
                except KeyError:
                    build_glyph_program(g, {}, xgf_doc)
                except Exception as e:
                    print("Exception in build_glyph_program:")
                    print(type(e))
                    print(e)
    return xgf_doc



from PyQt6.QtWidgets import QDialog, QVBoxLayout, QHBoxLayout, QDialogButtonBox, QComboBox, QLineEdit
from PyQt6.QtGui import QRegularExpressionValidator
from PyQt6.QtCore import QRegularExpression, pyqtSlot
import uharfbuzz as hb
from tempfile import SpooledTemporaryFile
from .freetypeFont import freetypeFont


class harfbuzzFont:
    LANG_DEFAULT = 0xFFFF

    SCRIPT_DEFAULT = 0

    DEFAULT_LAYOUT_TAGS = [
        "ccmp",
        "liga",
        "calt",
        "rlig",
        "locl"
    ]

    # We omit cvNN and ssNN tags, but instead construct these when needed.
    LAYOUT_TAGS = {
        "abvf": "Above-base Forms",
        "abvm": "Above-base Mark Positioning",
        "abvs": "Above-base Substitutions",
        "blwf": "Below-base Forms",
        "blwm": "Below-base Mark Positioning",
        "blws": "Below-base Substitutions",
        "pref": "Pre-base Forms",
        "pres": "Pre-base Substitutions",
        "psts": "Post-base Substitutions",
        "pstf": "Post-base Forms",
        "dist": "Distance",
        "akhn": "Akhand",
        "haln": "Halant Forms",
        "half": "Half Form",
        "nukt": "Nukta Forms",
        "rkrf": "Rakar Forms",
        "rphf": "Reph Form",
        "vatu": "Vattu Variants",
        "cjct": "Conjunct Forms",
        "cfar": "Conjunct Form After Ro",
        "smpl": "Simplified Forms",
        "trad": "Traditional Forms",
        "tnam": "Traditional Name Forms",
        "expt": "Expert Forms",
        "hojo": "Hojo Kanji Forms",
        "nlck": "NLC Kanji Forms",
        "jp78": "JIS 78 Forms",
        "jp83": "JIS 83 Forms",
        "jp90": "JIS 90 Forms",
        "jp04": "JIS 04 Forms",
        "hngl": "Hangul",
        "ljmo": "Leading Jamo Forms",
        "tjmo": "Trailing Jamo Forms",
        "vjmo": "Vowel Jamo Forms",
        "fwid": "Full Widths",
        "hwid": "Half Widths",
        "halt": "Alternate Half Widths",
        "twid": "Third Widths",
        "qwid": "Quarter Widths",
        "pwid": "Proportional Widths",
        "palt": "Proportional Alternates",
        "pkna": "Proportional Kana",
        "ruby": "Ruby Notation Forms",
        "hkna": "Horizontal Kana Alternates",
        "vkna": "Vertical Kana",
        "cpct": "Centered CJK Punctuation",
        "curs": "Cursive Positioning",
        "jalt": "Justification Alternates",
        "mset": "Mark Positioning via Substitution",
        "rclt": "Required Contextual Alternates",
        "rlig": "Required Ligatures",
        "isol": "Isolated Forms",
        "init": "Initial Forms",
        "medi": "Medial Forms",
        "med2": "Medial Form #2",
        "fina": "Terminal Forms",
        "fin2": "Terminal Form #2",
        "fin3": "Terminal Form #3",
        "falt": "Final Glyph on Line Alternates",
        "stch": "Stretching Glyph Decomposition",
        "smcp": "Small Caps",
        "c2sc": "Capitals to Small Caps",
        "pcap": "Petite Caps",
        "c2pc": "Capitals to Petite Caps",
        "unic": "Unicase",
        "cpsp": "Capital Spacing",
        "case": "Case Sensitive Forms",
        "ital": "Italics",
        "ordn": "Ordinals",
        "valt": "Alternative Vertical Metrics",
        "vhal": "Alternative Vertical Half Metrics",
        "vpal": "Proportional Alternate Vertical Metrics",
        "vert": "Vertical Alternates",
        "vrt2": "Vertical Alternates and Rotation",
        "vrtr": "Vertical Alternates for Rotation",
        "vkrn": "Vertical Kerning",
        "ltra": "Left-to-right glyph alternates",
        "ltrm": "Left-to-right mirrored forms",
        "rtla": "Right-to-left glyph alternates",
        "rtlm": "Right-to-left mirrored forms",
        "lnum": "Lining Figures",
        "onum": "Oldstyle Figures",
        "pnum": "Proportional Figures",
        "tnum": "Tabular Figures",
        "frac": "Fractions",
        "afrc": "Alternative Fractions",
        "dnom": "Denominator",
        "numr": "Numerator",
        "sinf": "Scientific Inferiors",
        "zero": "Slashed Zero",
        "mgrk": "Mathematical Greek",
        "flac": "Flattened accent forms",
        "dtls": "Dotless Forms",
        "ssty": "Math script style alternates",
        "aalt": "Access All Alternates",
        "swsh": "Swash",
        "cswh": "Contextual Swash",
        "calt": "Contextual Alternates",
        "hist": "Historical Forms",
        "locl": "Localized Forms",
        "rand": "Randomize",
        "nalt": "Alternate Annotation Forms",
        "salt": "Stylistic Alternates",
        "subs": "Subscript",
        "sups": "Superscript",
        "titl": "Titling Alternates",
        "rvrn": "Required Variation Alternates",
        "clig": "Contextual Ligatures",
        "dlig": "Discretionary Ligatures",
        "hlig": "Historical Ligatures",
        "liga": "Standard Ligatures",
        "ccmp": "Glyph Composition/Decomposition",
        "kern": "Kerning",
        "mark": "Mark Positioning",
        "mkmk": "Mark-to-mark Positioning",
        "opbd": "Optical Bounds",
        "lfbd": "Left Bounds",
        "rtbd": "Right Bounds",
        "ornm": "Ornaments",
    }

    SCRIPT_TAGS = {
        'adlm': "Adlam",
        'ahom': "Ahom",
        'hluw': "Anatolian Hieroglyphs",
        'arab': "Arabic",
        'armn': "Armenian",
        'avst': "Avestan",
        'bali': "Balinese",
        'bamu': "Bamum",
        'bass': "Bassa Vah",
        'batk': "Batak",
        'beng': "Bengali",
        'bng2': "Bengali v.2",
        'bhks': "Bhaiksuki",
        'bopo': "Bopomofo",
        'brah': "Brahmi",
        'brai': "Braille",
        'bugi': "Buginese",
        'buhd': "Buhid",
        'byzm': "Byzantine Music",
        'cans': "Canadian Syllabics",
        'cari': "Carian",
        'aghb': "Caucasian Albanian",
        'cakm': "Chakma",
        'cham': "Cham",
        'cher': "Cherokee",
        'chrs': "Chorasmian",
        'hani': "CJK Ideographic",
        'copt': "Coptic",
        'cprt': "Cypriot Syllabary",
        'cpmn': "Cypro-Minoan",
        'cyrl': "Cyrillic",
        'DFLT': "Default",
        'dsrt': "Deseret",
        'deva': "Devanagari",
        'dev2': "Devanagari v.2",
        'diak': "Dives Akuru",
        'dogr': "Dogra",
        'dupl': "Duployan",
        'egyp': "Egyptian Hieroglyphs",
        'elba': "Elbasan",
        'elym': "Elymaic",
        'ethi': "Ethiopic",
        'geor': "Georgian",
        'glag': "Glagolitic",
        'goth': "Gothic",
        'gran': "Grantha",
        'grek': "Greek",
        'gujr': "Gujarati",
        'gjr2': "Gujarati v.2",
        'gong': "Gunjala Gondi",
        'guru': "Gurmukhi",
        'gur2': "Gurmukhi v.2",
        'hang': "Hangul",
        'jamo': "Hangul Jamo",
        'rohg': "Hanifi Rohingya",
        'hano': "Hanunoo",
        'hatr': "Hatran",
        'hebr': "Hebrew",
        'kana': "Hiragana",
        'armi': "Imperial Aramaic",
        'phli': "Inscriptional Pahlavi",
        'prti': "Inscriptional Parthian",
        'java': "Javanese",
        'kthi': "Kaithi",
        'knda': "Kannada",
        'knd2': "Kannada v.2",
        'kana': "Katakana",
        'kali': "Kayah Li",
        'khar': "Kharosthi",
        'kits': "Khitan Small Script",
        'khmr': "Khmer",
        'khoj': "Khojki",
        'sind': "Khudawadi",
        'lao ': "Lao",
        'latn': "Latin",
        'lepc': "Lepcha",
        'limb': "Limbu",
        'lina': "Linear A",
        'linb': "Linear B",
        'lisu': "Lisu (Fraser)",
        'lyci': "Lycian",
        'lydi': "Lydian",
        'mahj': "Mahajani",
        'maka': "Makasar",
        'mlym': "Malayalam",
        'mlm2': "Malayalam v.2",
        'mand': "Mandaic, Mandaean",
        'mani': "Manichaean",
        'marc': "Marchen",
        'gonm': "Masaram Gondi",
        'math': "Mathematical Alphanumeric Symbols",
        'medf': "Medefaidrin (Oberi Okaime, Oberi Ɔkaimɛ)",
        'mtei': "Meitei Mayek (Meithei, Meetei)",
        'mend': "Mende Kikakui",
        'merc': "Meroitic Cursive",
        'mero': "Meroitic Hieroglyphs",
        'plrd': "Miao",
        'modi': "Modi",
        'mong': "Mongolian",
        'mroo': "Mro",
        'mult': "Multani",
        'musc': "Musical Symbols",
        'mymr': "Myanmar",
        'mym2': "Myanmar v.2",
        'nbat': "Nabataean",
        'nand': "Nandinagari",
        'newa': "Newa",
        'talu': "New Tai Lue",
        'nko ': "N'Ko",
        'nshu': "Nüshu",
        'hmnp': "Nyiakeng Puachue Hmong",
        'orya': "Odia (formerly Oriya)",
        'ory2': "Odia v.2 (formerly Oriya v.2)",
        'ogam': "Ogham",
        'olck': "Ol Chiki",
        'ital': "Old Italic",
        'hung': "Old Hungarian",
        'narb': "Old North Arabian",
        'perm': "Old Permic",
        'xpeo': "Old Persian Cuneiform",
        'sogo': "Old Sogdian",
        'sarb': "Old South Arabian",
        'orkh': "Old Turkic, Orkhon Runic",
        'ougr': "Old Uyghur",
        'osge': "Osage",
        'osma': "Osmanya",
        'hmng': "Pahawh Hmong",
        'palm': "Palmyrene",
        'pauc': "Pau Cin Hau",
        'phag': "Phags-pa",
        'phnx': "Phoenician",
        'phlp': "Psalter Pahlavi",
        'rjng': "Rejang",
        'runr': "Runic",
        'samr': "Samaritan",
        'saur': "Saurashtra",
        'shrd': "Sharada",
        'shaw': "Shavian",
        'sidd': "Siddham",
        'sgnw': "Sign Writing",
        'sinh': "Sinhala",
        'sogd': "Sogdian",
        'sora': "Sora Sompeng",
        'soyo': "Soyombo",
        'xsux': "Sumero-Akkadian Cuneiform",
        'sund': "Sundanese",
        'sylo': "Syloti Nagri",
        'syrc': "Syriac",
        'tglg': "Tagalog",
        'tagb': "Tagbanwa",
        'tale': "Tai Le",
        'lana': "Tai Tham (Lanna)",
        'tavt': "Tai Viet",
        'takr': "Takri",
        'taml': "Tamil",
        'tml2': "Tamil v.2",
        'tnsa': "Tangsa",
        'tang': "Tangut",
        'telu': "Telugu",
        'tel2': "Telugu v.2",
        'thaa': "Thaana",
        'thai': "Thai",
        'tibt': "Tibetan",
        'tfng': "Tifinagh",
        'tirh': "Tirhuta",
        'toto': "Toto",
        'ugar': "Ugaritic Cuneiform",
        'vai ': "Vai",
        'vith': "Vithkuqi",
        'wcho': "Wancho",
        'wara': "Warang Citi",
        'yezi': "Yezidi",
        'yi  ': "Yi",
        'zanb': "Zanabazar Square",
    }

    LANGUAGE_TAGS = {
        'ABA ': "Abaza",
        'ABK ': "Abkhazian",
        'ACH ': "Acholi",
        'ACR ': "Achi",
        'ADY ': "Adyghe",
        'AFK ': "Afrikaans",
        'AFR ': "Afar",
        'AGW ': "Agaw",
        'AIO ': "Aiton",
        'AKA ': "Akan",
        'AKB ': "Batak Angkola",
        'ALS ': "Alsatian",
        'ALT ': "Altai",
        'AMH ': "Amharic",
        'ANG ': "Anglo-Saxon",
        'ARA ': "Arabic",
        'ARG ': "Aragonese",
        'ARI ': "Aari",
        'ARK ': "Rakhine",
        'ASM ': "Assamese",
        'AST ': "Asturian",
        'ATH ': "Athapaskan languages",
        'AVN ': "Avatime",
        'AVR ': "Avar",
        'AWA ': "Awadhi",
        'AYM ': "Aymara",
        'AZB ': "Torki",
        'AZE ': "Azerbaijani",
        'BAD ': "Badaga",
        'BAG ': "Baghelkhandi",
        'BAL ': "Balkar",
        'BAN ': "Balinese",
        'BAR ': "Bavarian",
        'BAU ': "Baulé",
        'BBC ': "Batak Toba",
        'BBR ': "Berber",
        'BCH ': "Bench",
        'BCR ': "Bible Cree",
        'BDY ': "Bandjalang",
        'BEL ': "Belarussian",
        'BEM ': "Bemba",
        'BEN ': "Bengali",
        'BGC ': "Haryanvi",
        'BGQ ': "Bagri",
        'BGR ': "Bulgarian",
        'BHI ': "Bhili",
        'BHO ': "Bhojpuri",
        'BIK ': "Bikol",
        'BIL ': "Bilen",
        'BIS ': "Bislama",
        'BJJ ': "Kanauji",
        'BKF ': "Blackfoot",
        'BLI ': "Baluchi",
        'BLK ': "Pa’o Karen",
        'BLN ': "Balante",
        'BLT ': "Balti",
        'BMB ': "Bambara (Bamanankan)",
        'BML ': "Bamileke",
        'BOS ': "Bosnian",
        'BPY ': "Bishnupriya Manipuri",
        'BRE ': "Breton",
        'BRH ': "Brahui",
        'BRI ': "Braj Bhasha",
        'BRM ': "Burmese",
        'BRX ': "Bodo",
        'BSH ': "Bashkir",
        'BSK ': "Burushaski",
        'BTD ': "Batak Dairi (Pakpak)",
        'BTI ': "Beti",
        'BTK ': "Batak languages",
        'BTM ': "Batak Mandailing",
        'BTS ': "Batak Simalungun",
        'BTX ': "Batak Karo",
        'BTZ ': "Batak Alas-Kluet",
        'BUG ': "Bugis",
        'BYV ': "Medumba",
        'CAK ': "Kaqchikel",
        'CAT ': "Catalan",
        'CBK ': "Zamboanga Chavacano",
        'CEB ': "Cebuano",
        'CGG ': "Chiga",
        'CHA ': "Chamorro",
        'CHE ': "Chechen",
        'CHG ': "Chaha Gurage",
        'CHH ': "Chattisgarhi",
        'CHI ': "Chichewa (Chewa, Nyanja)",
        'CHK ': "Chukchi",
        'CHO ': "Choctaw",
        'CHP ': "Chipewyan",
        'CHR ': "Cherokee",
        'CHU ': "Chuvash",
        'CHY ': "Cheyenne",
        'CJA ': "Western Cham",
        'CJM ': "Eastern Cham",
        'CMR ': "Comorian",
        'COP ': "Coptic",
        'COR ': "Cornish",
        'COS ': "Corsican",
        'CPP ': "Creoles",
        'CRE ': "Cree",
        'CRR ': "Carrier",
        'CRT ': "Crimean Tatar",
        'CSB ': "Kashubian",
        'CSL ': "Church Slavonic",
        'CSY ': "Czech",
        'CTG ': "Chittagonian",
        'CTT ': "Wayanad Chetti",
        'CUK ': "San Blas Kuna",
        'DAG ': "Dagbani",
        'DAN ': "Danish",
        'DAR ': "Dargwa",
        'DAX ': "Dayi",
        'DCR ': "Woods Cree",
        'DEU ': "German",
        'DGO ': "Dogri (individual language)",
        'DGR ': "Dogri (macrolanguage)",
        'DHG ': "Dhangu",
        'DHV ': "Divehi (Dhivehi, Maldivian)",
        'DIQ ': "Dimli",
        'DIV ': "Divehi (Dhivehi, Maldivian)",
        'DJR ': "Zarma",
        'DNG ': "Dangme",
        'DNJ ': "Dan",
        'DNK ': "Dinka",
        'DRI ': "Dari",
        'DUJ ': "Dhuwal",
        'DUN ': "Dungan",
        'DZN ': "Dzongkha",
        'EBI ': "Ebira",
        'ECR ': "Eastern Cree",
        'EDO ': "Edo",
        'EFI ': "Efik",
        'ELL ': "Greek",
        'EMK ': "Eastern Maninkakan",
        'ENG ': "English",
        'ERZ ': "Erzya",
        'ESP ': "Spanish",
        'ESU ': "Central Yupik",
        'ETI ': "Estonian",
        'EUQ ': "Basque",
        'EVK ': "Evenki",
        'EVN ': "Even",
        'EWE ': "Ewe",
        'FAN ': "French Antillean",
        'FAR ': "Persian",
        'FAT ': "Fanti",
        'FIN ': "Finnish",
        'FJI ': "Fijian",
        'FLE ': "Dutch (Flemish)",
        'FMP ': "Fe’fe’",
        'FNE ': "Forest Enets",
        'FON ': "Fon",
        'FOS ': "Faroese",
        'FRA ': "French",
        'FRC ': "Cajun French",
        'FRI ': "Frisian",
        'FRL ': "Friulian",
        'FRP ': "Arpitan",
        'FTA ': "Futa",
        'FUL ': "Fulah",
        'FUV ': "Nigerian Fulfulde",
        'GAD ': "Ga",
        'GAE ': "Scottish Gaelic (Gaelic)",
        'GAG ': "Gagauz",
        'GAL ': "Galician",
        'GAR ': "Garshuni",
        'GAW ': "Garhwali",
        'GEZ ': "Geez",
        'GIH ': "Githabul",
        'GIL ': "Gilyak",
        'GKP ': "Kpelle (Guinea)",
        'GLK ': "Gilaki",
        'GMZ ': "Gumuz",
        'GNN ': "Gumatj",
        'GOG ': "Gogo",
        'GON ': "Gondi",
        'GRN ': "Greenlandic",
        'GRO ': "Garo",
        'GUA ': "Guarani",
        'GUC ': "Wayuu",
        'GUF ': "Gupapuyngu",
        'GUJ ': "Gujarati",
        'GUZ ': "Gusii",
        'HAI ': "Haitian (Haitian Creole)",
        'HAI0': "Haida",
        'HAL ': "Halam (Falam Chin)",
        'HAR ': "Harauti",
        'HAU ': "Hausa",
        'HAW ': "Hawaiian",
        'HAY ': "Haya",
        'HAZ ': "Hazaragi",
        'HBN ': "Hammer-Banna",
        'HEI ': "Heiltsuk",
        'HER ': "Herero",
        'HIL ': "Hiligaynon",
        'HIN ': "Hindi",
        'HMA ': "High Mari",
        'HMD ': "A-Hmao",
        'HMN ': "Hmong",
        'HMO ': "Hiri Motu",
        'HMZ ': "Hmong Shuat",
        'HND ': "Hindko",
        'HO  ': "Ho",
        'HRI ': "Harari",
        'HRV ': "Croatian",
        'HUN ': "Hungarian",
        'HYE ': "Armenian",
        'IBA ': "Iban",
        'IBB ': "Ibibio",
        'IBO ': "Igbo",
        'IDO ': "Ido",
        'IJO ': "Ijo languages",
        'ILE ': "Interlingue",
        'ILO ': "Ilokano",
        'INA ': "Interlingua",
        'IND ': "Indonesian",
        'ING ': "Ingush",
        'INU ': "Inuktitut",
        'INUK': "Nunavik Inuktitut",
        'IPK ': "Inupiat",
        'IRI ': "Irish",
        'IRT ': "Irish Traditional",
        'IRU ': "Irula",
        'ISL ': "Icelandic",
        'ISM ': "Inari Sami",
        'ITA ': "Italian",
        'IWR ': "Hebrew",
        'JAM ': "Jamaican Creole",
        'JAN ': "Japanese",
        'JAV ': "Javanese",
        'JBO ': "Lojban",
        'JCT ': "Krymchak",
        'JII ': "Yiddish",
        'JUD ': "Ladino",
        'JUL ': "Jula",
        'KAB ': "Kabardian",
        'KAC ': "Kachchi",
        'KAL ': "Kalenjin",
        'KAN ': "Kannada",
        'KAR ': "Karachay",
        'KAT ': "Georgian",
        'KAW ': "Kawi (Old Javanese)",
        'KAZ ': "Kazakh",
        'KDE ': "Makonde",
        'KEA ': "Kabuverdianu (Crioulo)",
        'KEB ': "Kebena",
        'KEK ': "Kekchi",
        'KGE ': "Khutsuri Georgian",
        'KHA ': "Khakass",
        'KHK ': "Khanty-Kazim",
        'KHM ': "Khmer",
        'KHS ': "Khanty-Shurishkar",
        'KHT ': "Khamti Shan",
        'KHV ': "Khanty-Vakhi",
        'KHW ': "Khowar",
        'KIK ': "Kikuyu (Gikuyu)",
        'KIR ': "Kirghiz (Kyrgyz)",
        'KIS ': "Kisii",
        'KIU ': "Kirmanjki",
        'KJD ': "Southern Kiwai",
        'KJP ': "Eastern Pwo Karen",
        'KJZ ': "Bumthangkha",
        'KKN ': "Kokni",
        'KLM ': "Kalmyk",
        'KMB ': "Kamba",
        'KMN ': "Kumaoni",
        'KMO ': "Komo",
        'KMS ': "Komso",
        'KMZ ': "Khorasani Turkic",
        'KNR ': "Kanuri",
        'KOD ': "Kodagu",
        'KOH ': "Korean Old Hangul",
        'KOK ': "Konkani",
        'KOM ': "Komi",
        'KON ': "Kikongo",
        'KOP ': "Komi-Permyak",
        'KOR ': "Korean",
        'KOS ': "Kosraean",
        'KOZ ': "Komi-Zyrian",
        'KPL ': "Kpelle",
        'KRI ': "Krio",
        'KRK ': "Karakalpak",
        'KRL ': "Karelian",
        'KRM ': "Karaim",
        'KRN ': "Karen",
        'KRT ': "Koorete",
        'KSH ': "Kashmiri",
        'KSI ': "Khasi",
        'KSM ': "Kildin Sami",
        'KSW ': "S’gaw Karen",
        'KUA ': "Kuanyama",
        'KUI ': "Kui",
        'KUL ': "Kulvi",
        'KUM ': "Kumyk",
        'KUR ': "Kurdish",
        'KUU ': "Kurukh",
        'KUY ': "Kuy",
        'KWK ': "Kwakʼwala",
        'KYK ': "Koryak",
        'KYU ': "Western Kayah",
        'LAD ': "Ladin",
        'LAH ': "Lahuli",
        'LAK ': "Lak",
        'LAM ': "Lambani",
        'LAO ': "Lao",
        'LAT ': "Latin",
        'LAZ ': "Laz",
        'LCR ': "L-Cree",
        'LDK ': "Ladakhi",
        'LEF ': "Lelemi",
        'LEZ ': "Lezgi",
        'LIJ ': "Ligurian",
        'LIM ': "Limburgish",
        'LIN ': "Lingala",
        'LIS ': "Lisu",
        'LJP ': "Lampung",
        'LKI ': "Laki",
        'LMA ': "Low Mari",
        'LMB ': "Limbu",
        'LMO ': "Lombard",
        'LMW ': "Lomwe",
        'LOM ': "Loma",
        'LPO ': "Lipo",
        'LRC ': "Luri",
        'LSB ': "Lower Sorbian",
        'LSM ': "Lule Sami",
        'LTH ': "Lithuanian",
        'LTZ ': "Luxembourgish",
        'LUA ': "Luba-Lulua",
        'LUB ': "Luba-Katanga",
        'LUG ': "Ganda",
        'LUH ': "Luyia",
        'LUO ': "Luo",
        'LVI ': "Latvian",
        'MAD ': "Madura",
        'MAG ': "Magahi",
        'MAH ': "Marshallese",
        'MAJ ': "Majang",
        'MAK ': "Makhuwa",
        'MAL ': "Malayalam",
        'MAM ': "Mam",
        'MAN ': "Mansi",
        'MAP ': "Mapudungun",
        'MAR ': "Marathi",
        'MAW ': "Marwari",
        'MBN ': "Mbundu",
        'MBO ': "Mbo",
        'MCH ': "Manchu",
        'MCR ': "Moose Cree",
        'MDE ': "Mende",
        'MDR ': "Mandar",
        'MEN ': "Me’en",
        'MER ': "Meru",
        'MFA ': "Pattani Malay",
        'MFE ': "Morisyen",
        'MIN ': "Minangkabau",
        'MIZ ': "Mizo",
        'MKD ': "Macedonian",
        'MKR ': "Makasar",
        'MKW ': "Kituba",
        'MLE ': "Male",
        'MLG ': "Malagasy",
        'MLN ': "Malinke",
        'MLR ': "Malayalam Reformed",
        'MLY ': "Malay",
        'MND ': "Mandinka",
        'MNG ': "Mongolian",
        'MNI ': "Manipuri",
        'MNK ': "Maninka",
        'MNX ': "Manx",
        'MOH ': "Mohawk",
        'MOK ': "Moksha",
        'MOL ': "Moldavian",
        'MON ': "Mon",
        'MONT': "Thailand Mon",
        'MOR ': "Moroccan",
        'MOS ': "Mossi",
        'MRI ': "Maori",
        'MTH ': "Maithili",
        'MTS ': "Maltese",
        'MUN ': "Mundari",
        'MUS ': "Muscogee",
        'MWL ': "Mirandese",
        'MWW ': "Hmong Daw",
        'MYN ': "Mayan",
        'MZN ': "Mazanderani",
        'NAG ': "Naga-Assamese",
        'NAH ': "Nahuatl",
        'NAN ': "Nanai",
        'NAP ': "Neapolitan",
        'NAS ': "Naskapi",
        'NAU ': "Nauruan",
        'NAV ': "Navajo",
        'NCR ': "N-Cree",
        'NDB ': "Ndebele",
        'NDC ': "Ndau",
        'NDG ': "Ndonga",
        'NDS ': "Low Saxon",
        'NEP ': "Nepali",
        'NEW ': "Newari",
        'NGA ': "Ngbaka",
        'NGR ': "Nagari",
        'NHC ': "Norway House Cree",
        'NIS ': "Nisi",
        'NIU ': "Niuean",
        'NKL ': "Nyankole",
        'NKO ': "N’Ko",
        'NLD ': "Dutch",
        'NOE ': "Nimadi",
        'NOG ': "Nogai",
        'NOR ': "Norwegian",
        'NOV ': "Novial",
        'NSM ': "Northern Sami",
        'NSO ': "Northern Sotho",
        'NTA ': "Northern Tai",
        'NTO ': "Esperanto",
        'NYM ': "Nyamwezi",
        'NYN ': "Norwegian Nynorsk (Nynorsk, Norwegian)",
        'NZA ': "Mbembe Tigon",
        'OCI ': "Occitan",
        'OCR ': "Oji-Cree",
        'OJB ': "Ojibway",
        'ORI ': "Odia (formerly Oriya)",
        'ORO ': "Oromo",
        'OSS ': "Ossetian",
        'PAA ': "Palestinian Aramaic",
        'PAG ': "Pangasinan",
        'PAL ': "Pali",
        'PAM ': "Pampangan",
        'PAN ': "Punjabi",
        'PAP ': "Palpa",
        'PAS ': "Pashto",
        'PAU ': "Palauan",
        'PCC ': "Bouyei",
        'PCD ': "Picard",
        'PDC ': "Pennsylvania German",
        'PGR ': "Polytonic Greek",
        'PHK ': "Phake",
        'PIH ': "Norfolk",
        'PIL ': "Filipino",
        'PLG ': "Palaung",
        'PLK ': "Polish",
        'PMS ': "Piemontese",
        'PNB ': "Western Panjabi",
        'POH ': "Pocomchi",
        'PON ': "Pohnpeian",
        'PRO ': "Provençal / Old Provençal",
        'PTG ': "Portuguese",
        'PWO ': "Western Pwo Karen",
        'QIN ': "Chin",
        'QUC ': "K’iche’",
        'QUH ': "Quechua (Bolivia)",
        'QUZ ': "Quechua",
        'QVI ': "Quechua (Ecuador)",
        'QWH ': "Quechua (Peru)",
        'RAJ ': "Rajasthani",
        'RAR ': "Rarotongan",
        'RBU ': "Russian Buriat",
        'RCR ': "R-Cree",
        'REJ ': "Rejang",
        'RIA ': "Riang",
        'RHG ': "Rohingya",
        'RIF ': "Tarifit",
        'RIT ': "Ritarungo",
        'RKW ': "Arakwal",
        'RMS ': "Romansh",
        'RMY ': "Vlax Romani",
        'ROM ': "Romanian",
        'ROY ': "Romany",
        'RSY ': "Rusyn",
        'RTM ': "Rotuman",
        'RUA ': "Kinyarwanda",
        'RUN ': "Rundi",
        'RUP ': "Aromanian",
        'RUS ': "Russian",
        'SAD ': "Sadri",
        'SAN ': "Sanskrit",
        'SAS ': "Sasak",
        'SAT ': "Santali",
        'SAY ': "Sayisi",
        'SCN ': "Sicilian",
        'SCO ': "Scots",
        'SCS ': "North Slavey",
        'SEK ': "Sekota",
        'SEL ': "Selkup",
        'SFM ': "Small Flowery Miao",
        'SGA ': "Old Irish",
        'SGO ': "Sango",
        'SGS ': "Samogitian",
        'SHI ': "Tachelhit",
        'SHN ': "Shan",
        'SIB ': "Sibe",
        'SID ': "Sidamo",
        'SIG ': "Silte Gurage",
        'SKS ': "Skolt Sami",
        'SKY ': "Slovak",
        'SLA ': "Slavey",
        'SLV ': "Slovenian",
        'SML ': "Somali",
        'SMO ': "Samoan",
        'SNA ': "Sena",
        'SND ': "Sindhi",
        'SNH ': "Sinhala (Sinhalese)",
        'SNK ': "Soninke",
        'SOG ': "Sodo Gurage",
        'SOP ': "Songe",
        'SOT ': "Southern Sotho",
        'SQI ': "Albanian",
        'SRB ': "Serbian",
        'SRD ': "Sardinian",
        'SRK ': "Saraiki",
        'SRR ': "Serer",
        'SSL ': "South Slavey",
        'SSM ': "Southern Sami",
        'STQ ': "Saterland Frisian",
        'SUK ': "Sukuma",
        'SUN ': "Sundanese",
        'SUR ': "Suri",
        'SVA ': "Svan",
        'SVE ': "Swedish",
        'SWA ': "Swadaya Aramaic",
        'SWK ': "Swahili",
        'SWZ ': "Swati",
        'SXT ': "Sutu",
        'SXU ': "Upper Saxon",
        'SYL ': "Sylheti",
        'SYR ': "Syriac",
        'Syre': "Syriac, Estrangela script-variant",
        'Syrj': "Syriac, Western script-variant",
        'Syrn': "Syriac, Eastern script-variant",
        'SZL ': "Silesian",
        'TAB ': "Tabasaran",
        'TAJ ': "Tajiki",
        'TAM ': "Tamil",
        'TAT ': "Tatar",
        'TCR ': "TH-Cree",
        'TDD ': "Dehong Dai",
        'TEL ': "Telugu",
        'TET ': "Tetum",
        'TGL ': "Tagalog",
        'TGN ': "Tongan",
        'TGR ': "Tigre",
        'TGY ': "Tigrinya",
        'THA ': "Thai",
        'THT ': "Tahitian",
        'TIB ': "Tibetan",
        'TIV ': "Tiv",
        'TJL ': "Tai Laing",
        'TKM ': "Turkmen",
        'TLI ': "Tlingit",
        'TMH ': "Tamashek",
        'TMN ': "Temne",
        'TNA ': "Tswana",
        'TNE ': "Tundra Enets",
        'TNG ': "Tonga",
        'TOD ': "Todo",
        'TPI ': "Tok Pisin",
        'TRK ': "Turkish",
        'TSG ': "Tsonga",
        'TSJ ': "Tshangla",
        'TUA ': "Turoyo Aramaic",
        'TUL ': "Tulu",
        'TUM ': "Tumbuka",
        'TUV ': "Tuvin",
        'TVL ': "Tuvalu",
        'TWI ': "Twi",
        'TYZ ': "Tày",
        'TZM ': "Tamazight",
        'TZO ': "Tzotzil",
        'UDM ': "Udmurt",
        'UKR ': "Ukrainian",
        'UMB ': "Umbundu",
        'URD ': "Urdu",
        'USB ': "Upper Sorbian",
        'UYG ': "Uyghur",
        'UZB ': "Uzbek",
        'VEC ': "Venetian",
        'VEN ': "Venda",
        'VIT ': "Vietnamese",
        'VOL ': "Volapük",
        'VRO ': "Võro",
        'WA  ': "Wa",
        'WAG ': "Wagdi",
        'WAR ': "Waray-Waray",
        'WCI ': "Waci Gbe",
        'WCR ': "West-Cree",
        'WEL ': "Welsh",
        'WLF ': "Wolof",
        'WLN ': "Walloon",
        'WTM ': "Mewati",
        'XBD ': "Lü",
        'XHS ': "Xhosa",
        'XJB ': "Minjangbal",
        'XKF ': "Khengkha",
        'XOG ': "Soga",
        'XPE ': "Kpelle (Liberia)",
        'XUB ': "Bette Kuruma",
        'XUJ ': "Jennu Kuruma",
        'YAK ': "Sakha",
        'YAO ': "Yao",
        'YAP ': "Yapese",
        'YBA ': "Yoruba",
        'YCR ': "Y-Cree",
        'YGP ': "Gepo",
        'YIC ': "Yi Classic",
        'YIM ': "Yi Modern",
        'YNA ': "Aluo",
        'YWQ ': "Wuding-Luquan Yi",
        'ZEA ': "Zealandic",
        'ZGH ': "Standard Moroccan Tamazight",
        'ZHA ': "Zhuang",
        'ZHH ': "Chinese, Traditional, Hong Kong SAR",
        'ZHP ': "Chinese, Phonetic",
        'ZHS ': "Chinese, Simplified",
        'ZHT ': "Chinese, Traditional",
        'ZHTM': "Chinese, Traditional, Macao SAR",
        'ZND ': "Zande",
        'ZUL ': "Zulu",
        'ZZA ': "Zazaki",
    }

    def __init__(
        self,
        font: SpooledTemporaryFile | str,
        ft_font: freetypeFont,
        keep_open=False,
    ):
        self.ft_font = ft_font

        #
        # Read the font
        #
        if type(font) is SpooledTemporaryFile:
            font.seek(0)
            font_data = font.read()
            if not keep_open:
                font.close()
        else:
            font_file = open(font, "rb")
            font_data = font_file.read()
            if not keep_open:
                font_file.close()
        self.hb_face = hb.Face(font_data)
        self.hb_font = hb.Font(self.hb_face)

        #
        # Get tag lists: script, language, features
        #
        self._pos_features = hb._harfbuzz.ot_layout_language_get_feature_tags(
            self.hb_face, "GPOS"
        )
        self._sub_scripts = hb.ot_layout_table_get_script_tags(self.hb_face, "GSUB")
        self._sub_languages = ["dflt"]
        self._sub_features = []
        self._active_features = {}
        self.current_script_tag = ""
        self.current_language_tag = ""
        if "DFLT" in self._sub_scripts:
            self.current_script_tag = "DFLT"
        else:
            if len(self._sub_scripts):
                self.current_script_tag = self._sub_scripts[0]
        if self.current_script_tag:
            self._sub_languages.extend(
                hb.ot_layout_script_get_language_tags(
                    self.hb_face, "GSUB", script_index=self.current_script_index
                )
            )
            self.current_language_tag = "dflt"
            self._sub_features = hb._harfbuzz.ot_layout_language_get_feature_tags(
                self.hb_face,
                "GSUB",
                script_index=self.current_script_index,
                language_index=self.current_language_index,
            )
            self._sub_features = sorted(self._sub_features)
            self.set_default_features()

    def set_default_features(self) -> None:
        self._active_features.clear()
        if "ccmp" in self._sub_features:
            self._active_features["ccmp"] = True
        if "liga" in self._sub_features:
            self._active_features["liga"] = True
        if "calt" in self._sub_features:
            self._active_features["calt"] = True
        if "rlig" in self._sub_features:
            self._active_features["rlig"] = True
        if "locl" in self._sub_features:
            self._active_features["locl"] = True
        if "abvm" in self._pos_features:
            self._active_features["abvm"] = True
        if "blwm" in self._pos_features:
            self._active_features["blwm"] = True
        if "clig" in self._sub_features:
            self._active_features["clig"] = True
        if "curs" in self._sub_features:
            self._active_features["curs"] = True
        if "dist" in self._sub_features:
            self._active_features["dist"] = True
        if "rclt" in self._sub_features:
            self._active_features["rclt"] = True
        if "kern" in self._pos_features:
            self._active_features["kern"] = True
        if "mark" in self._pos_features:
            self._active_features["mark"] = True
        if "mkmk" in self._pos_features:
            self._active_features["mkmk"] = True

    def set_coordinates(self, d: dict) -> None:
        self.hb_font.set_variations(d)

    @property
    def current_script_index(self) -> int:
        try:
            return self._sub_scripts.index(self.current_script_tag)
        except ValueError:
            return harfbuzzFont.SCRIPT_DEFAULT

    @property
    def current_language_index(self) -> int:
        try:
            ind = self._sub_languages.index(self.current_language_tag)
            if ind == 0:
                ind = harfbuzzFont.LANG_DEFAULT
            else:
                ind -= 1
        except ValueError:
            ind = harfbuzzFont.LANG_DEFAULT
        return ind

    @property
    def sub_features(self) -> list:
        return self._sub_features

    @property
    def sub_scripts(self) -> list:
        return self._sub_scripts

    @property
    def sub_languages(self) -> list:
        return self._sub_languages

    #@property
    #def pos_features(self) -> list:
    #    return self._pos_features

    #@property
    #def all_features(self) -> list:
    #    result = []
    #    result.extend(self.sub_features)
    #    result.extend(self.pos_features)
    #    return result

    @property
    def active_features(self) -> list:
        return self._active_features
    
    @classmethod
    def expanded_script_name(self, tag: str) -> str:
        try:
            return tag + " - " + harfbuzzFont.SCRIPT_TAGS[tag]
        except KeyError:
            return tag

    @classmethod
    def expanded_language_name(self, tag: str) -> str:
        try:
            return tag + " - " + harfbuzzFont.LANGUAGE_TAGS[tag]
        except KeyError:
            return tag

    @classmethod
    def expanded_feature_name(self, tag) -> str:
        ss = tag
        prefix = tag[0:2]
        try:
            if prefix == "cv":
                ss = tag + " - " + "Character Variant " + tag[2:4]
            elif prefix == "ss":
                ss = tag + " - " + "Stylistic Set " + tag[2:4]
            else:
                ss = tag + " - " + harfbuzzFont.LAYOUT_TAGS[tag]
        except Exception:
            pass
        return ss

    @classmethod
    def tag_only(self, s) -> str:
        tag = s.split(" ")[0]
        if len(tag) == 2:
            tag += "  "
        elif len(tag) == 3:
            tag += " "
        assert len(tag) == 4, "Length of layout tag must be 4 (2)."
        return tag

    def select_script(self, s: str) -> None:
        """ If the currently selected language is not available for the newly
            selected script, change language to 'dflt'.
        """
        if s in self._sub_scripts:
            self.current_script_tag = s
        self.select_language(self.current_language_tag)

    def select_language(self, l: str) -> None:
        """ New tag should have been selected from a list of available tags.
            And set up feature list for newly selected language.
        """
        self.current_language_tag = ""
        self._sub_languages.clear()
        self._sub_languages.append("dflt")
        self._sub_languages.extend(
            hb.ot_layout_script_get_language_tags(
                self.hb_face, "GSUB", script_index=self.current_script_index
            )
        )
        if l in self._sub_languages:
            self.current_language_tag = l
        else:
            self.current_language_tag = "dflt"
        self._sub_features.clear()
        self._sub_features.extend(hb._harfbuzz.ot_layout_language_get_feature_tags(
            self.hb_face,
            "GSUB",
            script_index=self.current_script_index,
            language_index=self.current_language_index,
        ))
        self._sub_features = sorted(self._sub_features)
        self.set_default_features()

    def activate_feature(self, f: str, index: int = -1) -> None:
        """ Activate an OT feature. if index > 0, that is included
            as well, for indexed features like salt.
        """
        add_feature = False
        if f in self._pos_features:
            add_feature = True
        val = True
        if f in self._sub_features:
            # If index == 0, that means don't add the feature.
            if index > 0:
                val = index
                add_feature = True
            elif index < 0:
                add_feature = True
        if add_feature and not f in self._active_features:
            self._active_features[f] = val

    def deactivate_feature(self, f: str) -> None:
        try:
            del self._active_features[f]
        except KeyError:
            pass

    def get_shaped_names(self, s: str):
        """ Run shape() (below) on string s, and return:
            1. A list of glyph names
            2. Hb's buf.glyph_positions (all the metrics data we need)
        """
        buf = self.hb_buffer(s)
        if self.current_script_tag:
            buf.script = self.current_script_tag
            if self.current_language_tag:
                buf.language = self.current_language_tag
        info, pos = self.shape(buf)
        indices = []
        for i in info:
            indices.append(i.codepoint)
        return self.ft_font.indices_to_names(indices), pos

    def shape(self, buf):
        """ Run hb.shape() on a Harfbuzz buffer and return buf.glyph_infos
            and buf.glyph_positions exactly as hb returns them.
        """
        buf.guess_segment_properties()
        hb.shape(self.hb_font, buf, self._active_features)
        return buf.glyph_infos, buf.glyph_positions

    def hb_buffer(self, s: str) -> hb.Buffer:
        """ Create an empty hb.Buffer and add string s to it.
        """
        buf = hb.Buffer.create()
        buf.add_str(s)
        return buf

    # @pyqtSlot()
    def reset_features(self):
        self.set_default_features()


class hbFeatureDialog(QDialog):
    def __init__(self, top_window, hb_font):
        super().__init__()
        self.valid = True
        self.top_window = top_window
        self.hb_font = hb_font
        self.setWindowTitle("Set an OpenType feature")
        self._layout = QVBoxLayout()
        self.input_layout = QHBoxLayout()
        QBtn = (
            QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        )
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.feature_list = []
        for s in self.hb_font._sub_features:
            self.feature_list.append(harfbuzzFont.expanded_feature_name(s))
        self.feature_list = sorted(self.feature_list)
        if len(self.feature_list):
            self.feature_box = QComboBox()
            for f in self.feature_list:
                self.feature_box.addItem(f)
            self.line_editor = QLineEdit()
            validator = QRegularExpressionValidator()
            re = QRegularExpression("\\b(on|off|[1-9]|[1-9][0-9])\\b")
            validator.setRegularExpression(re)
            self.line_editor.setValidator(validator)
            self.input_layout.addWidget(self.feature_box)
            self.input_layout.addWidget(self.line_editor)
            self._layout.addLayout(self.input_layout)
            self._layout.addWidget(self.buttonBox)
            self.setLayout(self._layout)
        else:
            self.valid = False

    @pyqtSlot()
    def accept(self) -> None:
        t = self.line_editor.text()
        current_tag = harfbuzzFont.tag_only(self.feature_box.currentText())
        val = None
        if len(t):
            try:
                val = int(t)
            except ValueError:
                if t == "on":
                    val = True
                elif t == "off":
                    val = False
        if val != None:
            self.hb_font.active_features[current_tag] = val
        super().accept()



from typing import Optional, Callable, Any
from PyQt6.QtCore import pyqtSignal, Qt, QTimer, pyqtSlot
from PyQt6.QtWidgets import QDialog, QVBoxLayout, QPlainTextEdit, QDialogButtonBox
from PyQt6.QtGui import QSyntaxHighlighter, QTextCharFormat, QColor
import yaml
import re
from yaml import Dumper
import copy
from schema import SchemaError # type: ignore
from .ygSchema import is_valid, set_error_message, error_message, have_error_message
from .ygModel import ygSourceable
from .ygPreferences import ygPreferences


# From https://stackoverflow.com/questions/8640959/
# how-can-i-control-what-scalar-form-pyyaml-uses-for-my-data
# Presumed public domain, since it was posted in a public forum.
def str_presenter(dumper, data):
    if len(data.splitlines()) > 1:  # check for multiline string
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


yaml.add_representer(str, str_presenter)

# to use with safe_dump:
yaml.representer.SafeRepresenter.add_representer(str, str_presenter)


class ygYAMLEditor(QPlainTextEdit):
    """An editor for source code for the current axis of the current glyph.

    Params:

    preferences (ygPreferences): The preferences object for the current file.
    """

    sig_source_from_editor = pyqtSignal(object)
    sig_status = pyqtSignal(object)
    sig_error = pyqtSignal(object)

    def __init__(self, preferences, parent = None) -> None:
        super().__init__()
        self.setAttribute(Qt.WidgetAttribute.WA_AcceptTouchEvents, False)
        self.setStyleSheet(
            "ygYAMLEditor {font-family: Source Code Pro, monospace; }"
        )
        # ; background-color: white
        self.setLineWrapMode(QPlainTextEdit.LineWrapMode.NoWrap)
        self.preferences = preferences
        self.textChanged.connect(self.text_changed)
        self._highlighter = ygGlyphHighlighter()
        self._timer = QTimer()
        self._timer.timeout.connect(self.check_valid)
        self.code_valid = True
        self.setup_editor()

    def setup_error_signal(self, f: Callable) -> None:
        self.sig_error.connect(f)

    @pyqtSlot(object)
    def install_source(self, l: list) -> None:
        self.setPlainText(l[0])
        if l[1]:
            self.setEnabled(False)
            return
        self.setEnabled(True)

    @pyqtSlot()
    def yaml_source(self) -> None:
        err = False
        # msg = ""
        s = ""
        try:
            s = yaml.safe_load(self.toPlainText())
        except Exception as e:
            err = True
            set_error_message("Source can't be parsed.")
            msg = str(e)
        if not err:
            try:
                err = not is_valid({"points": s})
            except SchemaError as se:
                err = True
        if err:
            self.sig_error.emit({"msg": error_message(), "mode": "console"})
        else:
            self.sig_source_from_editor.emit(s)

    def setup_status_indicator(self, o: Callable) -> None:
        self.sig_status.connect(o)

    def setup_editor_signals(self, f: Callable) -> None:
        self.sig_source_from_editor.connect(f)

    def disconnect_editor_signals(self, f: Callable) -> None:
        self.sig_source_from_editor.disconnect(f)

    @pyqtSlot()
    def check_valid(self) -> None:
        if not self.code_valid:
            if not have_error_message():
                set_error_message("Source can't be parsed.")
            self.sig_error.emit({"msg": error_message(), "mode": "console"})
            self.sig_status.emit(self.code_valid)

    @pyqtSlot()
    def text_changed(self) -> None:
        self.code_valid = True
        y = self.toPlainText()
        if len(y) == 0:
            self.setPlainText("[]\n")
        else:
            try:
                y = yaml.safe_load(y)
            except Exception as e:
                self.code_valid = False
            if self.code_valid:
                self.code_valid = is_valid({"points": y})
        # If code is not valid, start timer. Any time user presses a key,
        # the timer will restart if code is not (yet) valid. The effect is
        # that user has two seconds after any keypress to achieve validity
        # before any message is displayed.
        if self.code_valid:
            try:
                self._timer.stop()
            except Exception:
                pass
            self.sig_status.emit(self.code_valid)
        else:
            self._timer.start(2000)

    def setup_editor(self) -> None:
        tags = r"\b(ptid|ref|rel|macro|function|pos|dist|points|round|min)\:"
        twospace = r"(  |\- )"
        fourspace = r"(    |  \- )"
        sixspace = r"(      |    \- )"
        eightspace = r"(        |      \- )"
        tenspace = r"(          |        \- )"

        keytwo_format = QTextCharFormat()
        keytwo_format.setForeground(QColor(182, 6, 70, 255))
        pattern = twospace + tags
        self._highlighter.add_mapping(pattern, keytwo_format)

        keyfour_format = QTextCharFormat()
        keyfour_format.setForeground(QColor(75, 75, 255, 255))
        pattern = fourspace + tags
        self._highlighter.add_mapping(pattern, keyfour_format)

        keysix_format = QTextCharFormat()
        keysix_format.setForeground(QColor(201, 91, 12, 255))
        pattern = sixspace + tags
        self._highlighter.add_mapping(pattern, keysix_format)

        keyeight_format = QTextCharFormat()
        keyeight_format.setForeground(QColor("green"))
        pattern = eightspace + tags
        self._highlighter.add_mapping(pattern, keyeight_format)

        keyten_format = QTextCharFormat()
        keyten_format.setForeground(QColor("brown"))
        pattern = tenspace + tags
        self._highlighter.add_mapping(pattern, keyten_format)

        list_format = QTextCharFormat()
        list_format.setForeground(QColor(255, 75, 75, 255))
        pattern = r"^(\- |  \- |    \- |      \- |        \- |          \- )"
        self._highlighter.add_mapping(pattern, list_format)

        self._highlighter.setDocument(self.document())


class editorPane(QPlainTextEdit):
    """An editor for any chunk of code from current file, e.g. functions.
    This validates as the user types, but it only emits an error two
    seconds after user has stopped typing. This cuts back on unnecessary
    messages when (for example) a user is typing "true"--which is not
    valid until the word is complete.

    Params:

    owner: The dialog or window that owns this pane.

    sourceable: Object of type Sourceable: the thing to edit.

    validator: function that will throw an exception if text is not valid.

    save_on_focus_out (bool): Whether to auto-save if user leaves this
    editor.

    """

    sig_error = pyqtSignal(object)

    def __init__(
            self,
            owner: "editorDialog",
            sourceable: ygSourceable,
            validator: Callable,
            save_on_focus_out: bool = False
        ) -> None:
        super().__init__()
        self.save_on_focus_out = save_on_focus_out
        self.owner = owner
        self.textChanged.connect(self.text_changed)
        # error_state is true if the code in this editor is invalid.
        self.error_state = False
        self.set_style()
        self.watching_for_changes = False
        self.is_valid = validator
        self.sourceable = sourceable
        self.install_yaml(copy.copy(self.sourceable.source()))
        self.dirty = False
        self._timer = QTimer()
        self._timer.timeout.connect(self.check_valid)

    def setup_error_signal(self, f: Callable):
        self.sig_error.connect(f)

    def install_text(self, text: str) -> None:
        self.setPlainText(text)
        self.watching_for_changes = True

    def install_yaml(self, y: Any) -> None:
        try:
            t = yaml.dump(y, sort_keys=False, Dumper=Dumper)
        except Exception as e:
            print(e)
            t = self.owner._empty_string
        self.install_text(t)

    def set_style(self) -> None:
        if self.error_state:
            self.setStyleSheet(
                "QPlainTextEdit {font-family: Source Code Pro, monospace; color: black; background-color: rgb(252,227,242);  }"
            )
        else:
            self.setStyleSheet(
                "QPlainTextEdit {font-family: Source Code Pro, monospace; color: black; background-color: white;  }"
            )

    def set_error_state(self, b: bool) -> None:
        if self.error_state != b:
            self.error_state = b
            self.set_style()

    def yaml_source(self) -> Any:
        try:
            t = yaml.safe_load(self.toPlainText())
            self.set_error_state(False)
        except Exception as e:
            self.set_error_state(True)
            self.sig_error.emit({"msg": "Source can't be parsed.", "mode": "console"})
            t = self.owner._empty_string
        return t

    @pyqtSlot()
    def check_valid(self) -> None:
        self.set_error_state(True)
        self.sig_error.emit({"msg": error_message(), "mode": "console"})

    @pyqtSlot()
    def text_changed(self) -> None:
        if not self.watching_for_changes:
            return
        self.sourceable.set_clean(False)
        if len(self.toPlainText()) == 0:
            self.setPlainText(self.owner._empty_string)
        v = False
        try:
            v = self.is_valid(yaml.safe_load(self.toPlainText()))
        except Exception as e:
            set_error_message("Source can't be parsed.")
        if v:
            self._timer.stop()
            self.set_error_state(not v)
            self.owner.set_dialog_title(True)
        else:
            self._timer.start(2000)
            self.owner.set_dialog_title(False)
        self.dirty = True

    def focusOutEvent(self, event) -> None:
        if self.save_on_focus_out and self.dirty:
            c = self.yaml_source()
            if c != None:
                if self.is_valid(c):
                    self.sourceable.save(c)
                    self.set_error_state(False)
                else:
                    self.set_error_state(True)
            else:
                self.set_error_state(True)
                set_error_message("Source can't be parsed.")
            if self.error_state:
                self.sig_error.emit({"msg": error_message(), "mode": "console"})

    def showEvent(self, event) -> None:
        self.install_yaml(copy.copy(self.sourceable.source()))

    def refresh(self) -> None:
        self.install_yaml(copy.copy(self.sourceable.source()))


class editorDialog(QDialog):
    def __init__(
            self,
            preferences: ygPreferences,
            sourceable: ygSourceable,
            title: str,
            validator: Callable,
            top_structure: str = "dict"
        ) -> None:
        super().__init__()
        self.title = title
        self.set_dialog_title(True)
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        self.preferences = preferences
        self.sourceable = sourceable
        if top_structure == "dict":
            self._empty_string = "{}\n"
        else:
            self._empty_string = "[]\n"
        self.setMinimumSize(500, 500)
        self._layout = QVBoxLayout()
        self.setLayout(self._layout)
        self.edit_pane = editorPane(self, sourceable, validator)
        self.edit_pane.setup_error_signal(
            self.preferences.top_window().error_manager.new_message
        )
        self._layout.addWidget(self.edit_pane)
        QBtn = (
            QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        )
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self._layout.addWidget(self.buttonBox)

    def set_dialog_title(self, is_valid: bool) -> None:
        v = " (Invalid)"
        if is_valid:
            v = " (Valid)"
        self.setWindowTitle(self.title + v)

    def reject(self) -> None:
        self.done(QDialog.DialogCode.Rejected)

    def accept(self) -> None:
        err = False
        c = self.edit_pane.yaml_source()
        if c != None:
            if self.edit_pane.is_valid(c):
                self.sourceable.save(c)
            else:
                err = True
        else:
            err = True
        if err:
            self.reject()
            return
        self.done(QDialog.DialogCode.Accepted)


class ygGlyphHighlighter(QSyntaxHighlighter):
    def __init__(self, parent = None) -> None:
        QSyntaxHighlighter.__init__(self, parent)
        self._mappings: dict = {}

    def add_mapping(self, pattern: str, format: QTextCharFormat) -> None:
        self._mappings[pattern] = format

    def highlightBlock(self, text: str) -> None:
        for pattern, format in self._mappings.items():
            matches = re.finditer(pattern, text)
            for match in matches:
                s = match.group(0)
                start = match.start()
                end = match.end()
                self.setFormat(start, end - start, format)

class ygDeleteGlyphProgramsDialog(QDialog):
    def __init__(self, parent = None) -> None:
        super().__init__()
        self.result = ""
        self.ed_pane = QPlainTextEdit()
        self.setWindowTitle("Delete Glyph Programs")
        self._layout = QVBoxLayout()
        self._layout.addWidget(self.ed_pane)
        QBtn = (
            QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        )
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self._layout.addWidget(self.buttonBox)
        self.setLayout(self._layout)

    def reject(self) -> None:
        self.done(QDialog.DialogCode.Rejected)

    def accept(self) -> None:
        self.result = self.ed_pane.toPlainText()
        self.done(QDialog.DialogCode.Accepted)


from typing import Any, Optional
# from collections import deque
from PyQt6.QtWidgets import QPlainTextEdit
# from PyQt6.QtCore import Qt, QSize, QThread, pyqtSlot, pyqtSignal, QObject, QEvent


class ygErrorWindow(QPlainTextEdit):
    def __init__(self, init_text: list = []) -> None:
        super().__init__()
        self.setMaximumBlockCount(100)
        self.setReadOnly(True)
        if len(init_text) > 0:
            self.fill_in(init_text)
        self.window().setWindowTitle("Error Console")

    def fill_in(self, msg_list: list) -> None:
        self.clear()
        for m in msg_list:
            self.appendPlainText(m)
        self.centerCursor()

    def add_message(self, m: str) -> None:
        self.appendPlainText("")
        self.appendPlainText(m)
        self.centerCursor()

    def closeEvent(self, event) -> None:
        self.hide()


class ygErrorMessages:
    def __init__(self, top_window: Any) -> None:
        self.error_pane: Optional[ygErrorWindow] = None
        self.top_window = top_window
        self.last_message = ""

    # @pyqtSlot(object)
    def new_message(self, m: dict) -> None:
        msg = m["msg"]
        if msg and (msg != self.last_message):
            self.last_message = msg
            if m["mode"] == "console":
                if not self.error_pane:
                    self.error_pane = ygErrorWindow()
                    qg = self.top_window.screen().availableGeometry()
                    x = qg.x() + 100
                    y = qg.y() + 100
                    width = qg.width() * 0.35
                    height = qg.height() * 0.30
                    self.error_pane.setGeometry(int(x), int(y), int(width), int(height))
                    self.error_pane.show()
                else:
                    if not self.error_pane.isVisible():
                        self.error_pane.show()
                        self.error_pane.raise_()
                self.error_pane.add_message(msg)
            else:
                self.top_window.show_error_message(["Error", "Error", msg])



from typing import Optional
from fontTools.varLib import instancer  # type: ignore
from fontTools.ttLib.ttFont import TTFont # type: ignore


class instanceChecker:
    """This class will build an instance for a variable font and check
    the y positions of points from which the cvt table was made. If
    the y positions are different, the y position of the point is
    recorded and reported.
    """

    def __init__(self, ft_font, cvt, masters) -> None:
        self.ft_font = ft_font
        self.cvt = cvt
        self.masters = masters
        self.axes = self.masters.yg_font.axes
        self.current_instance: Optional[TTFont] = None

    def refresh(self) -> None:
        self.delete_all_vars()
        d = self.get_all_variant_cvs()
        self.add_variants_to_cvt(d)

    def delete_all_vars(self) -> None:
        k = self.cvt.keys
        for kk in k:
            cv = self.cvt.get_cv(kk)
            try:
                # if there's no "origin" in the cv, we can't replace the vars.
                if "origin" in cv:
                    del cv["var"]
            except Exception:
                pass

    def make_instance(self, vals: dict) -> None:
        self.current_instance = instancer.instantiateVariableFont(self.ft_font, vals)

    def get_all_variant_cvs(self) -> dict:
        # We end up with a dict:
        # {master_id: {cv_name: val, ...}, ...}
        result = {}
        k = self.masters.keys
        for kk in k:
            c = self.get_cvs_for_master(kk)
            if len(c) > 0:
                result[kk] = c
        return result

    def add_variants_to_cvt(self, d: dict) -> None:
        # d is a dict in the format produced by get_all_variant_cvs.
        ck = self.cvt.keys
        dk = d.keys()
        for ckk in ck:  # iterate through CVs.
            for dkk in dk:  # iterate through masters
                if ckk in d[dkk]:  # if the master has an entry for the current CV,
                    cv = self.cvt.get_cv(ckk)  # create a "var" entry in the CV.
                    if not "var" in cv:
                        cv["var"] = {}
                    cv["var"][dkk] = d[dkk][ckk]

    def get_cvs_for_master(self, master_id: str) -> dict:
        # Get the glyph name and point index (or indices) from the
        # "origin" field. Get the list of y values (function below) and
        # use those to figure out the value of the cv for this glyph.
        # Compare the new position with that of the default cv.
        # If different, store. If the same, discard.
        result = {}
        coords = self.masters.get_master_coords(master_id)
        self.make_instance(coords)
        k = self.cvt.keys
        for kk in k:
            cv = self.cvt.get_cv(kk)
            if type(cv) is dict and "origin" in cv:
                if cv["type"] == "pos":
                    glyph_name = cv["origin"]["glyph"]
                    ptnum = cv["origin"]["ptnum"]
                    y_pos = self.y_list(glyph_name)[ptnum[0]]
                    if cv["val"] != y_pos:
                        result[kk] = y_pos
                elif cv["type"] == "dist":
                    glyph_name = cv["origin"]["glyph"]
                    ptnum = cv["origin"]["ptnum"]
                    yl = self.y_list(glyph_name)
                    y_diff = abs(yl[ptnum[0]] - yl[ptnum[1]])
                    if cv["val"] != y_diff:
                        result[kk] = y_diff
        return result

    def y_list(self, glyph_name: str) -> list:
        gl = self.current_instance["glyf"][glyph_name].getCoordinates(
            self.ft_font["glyf"][glyph_name]
        )
        y_points = []
        for p in gl[0]:
            y_points.append(p[1])
        return y_points



from typing import Optional, List, TYPE_CHECKING
from .ygModel import ygPoint, ygGlyph


class stemFinder:
    def __init__(self, p1: ygPoint, p2: ygPoint, yg_glyph: ygGlyph) -> None:
        self.yg_glyph = yg_glyph
        self.counter_clockwise = bool(self.yg_glyph.yg_font.defaults.get_default("counterclockwise"))
        self.contours = []
        contour = []
        points = yg_glyph.points
        for p in points:
            contour.append(p)
            if p.end_of_contour:
                self.contours.append(contour)
                contour = []
        self.high_point = p1
        self.low_point = p2
        if self.yg_glyph.axis == "y":
            # self.high_point must have a higher y value than self.low_point.
            if self.high_point.font_y < self.low_point.font_y:
                self.high_point, self.low_point = self.low_point, self.high_point
        else:
            # self.high_point must have a lower x value than self.low_point.
            if self.high_point.font_x > self.low_point.font_x:
                self.high_point, self.low_point = self.low_point, self.high_point

    def find_point_by_index(self, i: int, c: List[ygPoint]) -> Optional[ygPoint]:
        """ Find the point with index i in contour c. Returns None if not found.
        """
        for p in c:
            if p.index == i:
                return p
        return None

    def next_point(self, p: ygPoint, c: List[ygPoint]) -> ygPoint:
        """ p a ygPoint object; c a contour. Returns the next point
            on the contour, wrapping if necessary.
        """
        last_point = c[-1]
        first_point = c[0]
        if p.index < last_point.index:
            return self.find_point_by_index(p.index + 1, c)
        else:
            return first_point
        
    def prev_point(self, p: ygPoint, c: List[ygPoint]) -> ygPoint:
        last_point = c[-1]
        first_point = c[0]
        if p.index > first_point.index:
            return self.find_point_by_index(p.index - 1, c)
        else:
            return last_point

    def which_contour(self, pt: ygPoint) -> Optional[List[ygPoint]]:
        """ Return the contour (list of ygPoint objects) containing pt.
        """
        for c in self.contours:
            if pt in c:
                return c
        return None
    
    def x_direction(self, pt: ygPoint) -> str:
        """ Find the x direction of a line or curve at the location of
            point pt. Returns "left," "right," or "same" if this pt
            has the same x location as the next pt.

            To do: if the result is going to be "same," try with a
            prev_point function (to be supplied).
        """
        contour = self.which_contour(pt)
        next_point = self.next_point(pt, contour)
        this_x = pt.font_x
        next_x = next_point.font_x
        if next_x > this_x:
            if self.counter_clockwise:
                return("left")
            else:
                return "right"
        elif this_x > next_x:
            if self.counter_clockwise:
                return "right"
            else:
                return "left"
        # If the points are aligned on this axis we're probably at the end of a stem.
        prev_point = self.prev_point(pt, contour)
        prev_x = prev_point.font_x
        if prev_x < this_x:
            if self.counter_clockwise:
                return("left")
            else:
                return "right"
        elif this_x < prev_x:
            if self.counter_clockwise:
                return "right"
            else:
                return "left"
        return "same"

        
    def y_direction(self, pt: ygPoint) -> str:
        """ Find the y direction of a line or curve at the location of
            point pt. Returns "up," "down," or "same" if this pt
            has the same y location as the next pt.
        """
        contour = self.which_contour(pt)
        next_point = self.next_point(pt, contour)
        next_y = next_point.font_y
        this_y = pt.font_y
        if next_y > this_y:
            if self.counter_clockwise:
                return "down"
            else:
                return "up"
        elif this_y > next_y:
            if self.counter_clockwise:
                return "up"
            else:
                return "down"
        prev_point = self.prev_point(pt, contour)
        prev_y = prev_point.font_y
        if prev_y > this_y:
            if self.counter_clockwise:
                return "up"
            else:
                return "down"
        elif this_y > prev_y:
            if self.counter_clockwise:
                return "down"
            else:
                return "up"
        return "same"
        
    def get_color(self) -> str:
        """ Recommends a distance type for the stem formed by self.high_point
            and self.low_point, based on this class's analysis of the stem.
        """
        result = "graydist"
        if self.yg_glyph.axis == "x":
            high_y_dir = self.y_direction(self.high_point)
            low_y_dir = self.y_direction(self.low_point)
            if high_y_dir == "up" and low_y_dir == "down":
                result = "blackdist"
            elif high_y_dir == "down" and low_y_dir == "up":
                result = "whitedist"
        else:
            high_x_dir = self.x_direction(self.high_point)
            low_x_dir  = self.x_direction(self.low_point)
            if high_x_dir == "right" and low_x_dir == "left":
                result = "blackdist"
            elif high_x_dir == "left" and low_x_dir == "right":
                result = "whitedist"
        return result


# class horizontalSegmentList:
#     def __init__(self, c: List[ygPoint]):
#         self.pt_list = []
#         found_points = []
#         for p in c: